"""
Defines tools for dealing with functional groups and their reactions.

See the documentation of :class:`Reactor` to see how reactions between
functional groups are performed.

.. _`adding functional groups`:

Extending stk: Adding  more functional groups.
----------------------------------------------

If ``stk`` is to incorporate a new functional group, a new
:class:`FGInfo` instance should be added to
:data:`functional_groups`, which is defined in this module.

Adding a new :class:`FGInfo` instance to :data:`functional_groups` will
allow :meth:`.Topology.build` to connect the functional group to
all others during assembly. In most cases, nothing except adding this
instance should be necessary in order to incorporate new functional
groups.

Note that when adding SMARTS, if you want to make a SMARTS that targets
an atom in an environment, for example, a bromine connected to a
carbon::

    [$([Br][C])]

The atom you are targeting needs to be written first. The above SMARTS
works but::

    [$([C][Br])]

does not.

If a new functional group is to connect to another functional group
with a bond other than a single, the names of the functional groups
should be added to :attr:`Reactor.bond_orders`, along with the desired
bond order.

.. _`adding complex reactions`:

Extending stk: Adding complex reactions.
----------------------------------------

See :class:`Reactor`.

"""

from functools import partial
import numpy as np
from scipy.spatial.distance import euclidean
import rdkit.Chem.AllChem as rdkit
import rdkit.Geometry.rdGeometry as rdkit_geo
from collections import Counter
from ..utilities import AtomicPeriodicBond, flatten


class ReactionKey:
    """
    Used to create a key from a :class:`list` of fg names.

    In effect, this creates a unique id for each reaction, given a
    :class:`list` of functional groups involved in that reaction.

    Attributes
    ----------
    key : :class:`tuple`
        A unique key representing a reaction.

    """

    def __init__(self, *fg_names):
        """
        Intializer.

        Paramters
        ---------
        *fg_names : :class:`str`
            The names of functional groups involved in a reaction.

        """

        c = Counter(fg_names)
        self.key = tuple(
            sorted((key, value) for key, value in c.items())
        )

    def __eq__(self, other):
        return self.key == other.key

    def __hash__(self):
        return hash(self.key)

    def __repr__(self):
        fg_names = ', '.join(repr(name) for name, count in self.key
                             for i in range(count))

        return f'FGInfo({fg_names})'

    def __str__(self):
        return repr(self)


class Match:
    """
    A container for SMARTS queries.

    Attributes
    ----------
    smarts : :class:`str`
        A SMARTS string which matches some atoms.

    n : :class:`int`
        The maximum number of atoms to be matched by :attr:`smarts`,
        per functional group.

    """

    __slots__ = ['smarts', 'n']

    def __init__(self, smarts, n):
        self.smarts = smarts
        self.n = n

    def __eq__(self, other):
        return self.smarts == other.smarts and self.n == other.n

    def __repr__(self):
        return f'Match({self.smarts!r}, {self.n!r})'

    def __str__(self):
        return repr(self)


class FGInfo:
    """
    Contains key information about functional groups.

    The point of this class is to register which atoms of a functional
    group form bonds, and which are deleted during assembly of
    macromolecules.

    Attributes
    ----------
    name : :class:`str`
        The name of the functional group.

    fg_smarts : :class:`str`
        A SMARTS string which matches the functional group.

    bonder_smarts : :class:`list`
        A :class:`list` of the form

        .. code-block:: python

            bonder_smarts = [Match(smarts='[$([N]([H])[H])]', n=1),
                             Match(smarts='[$([H][N][H])]', n=1)]

        Each string is SMARTS string which matches a bonder atom in the
        functional group. The number represents how many matched atoms
        should be used as bonders, per functional group.

        In the example, ``Match(smarts='[$([N]([H])[H])]', n=1)``
        matches the nitrogen atom in the amine functional group. The
        ``n=1`` means that 1 nitrogen atom per functional group will be
        used as a bonder. The second
        ``Match(smarts='[$([H][N][H])]', n=1)``, matches the hydrogen
        atom in the amine functional group. Because ``n=1``, only 1 of
        hydrogen atoms per amine functional group will be used as a
        bonder. If instead ``Match(smarts='[$([H][N][H])]', n=2)`` was
        used, then both of the hydrogen atoms in the functional group
        would be used as bonders.

    del_smarts : :class:`list`
        Same as :attr:`bonder_smarts` but matched atoms are deleters

    """

    __slots__ = ['name', 'fg_smarts', 'bonder_smarts', 'del_smarts']

    def __init__(self, name, fg_smarts, bonder_smarts, del_smarts):
        """
        Initializes a :class:`FGInfo` instnace.

        Parameters
        ---------
        name : :class:`str`
            The name of the functional group.

        fg_smarts : :class:`str`
            A SMARTS string which matches the functional group.

        bonder_smarts : :class:`list`
            See :attr:`bonder_smarts`.

        del_smarts : :class:`list`
            See :attr:`del_smarts`.

        """

        self.name = name
        self.fg_smarts = fg_smarts
        self.bonder_smarts = bonder_smarts
        self.del_smarts = del_smarts

    def __eq__(self, other):
        return (self.name == other.name and
                self.fg_smarts == other.fg_smarts and
                self.bonder_smarts == other.bonder_smarts and
                self.del_smarts == other.del_smarts)

    def __repr__(self):
        return (f'FGInfo(name={self.name!r}, '
                f'fg_smarts={self.fg_smarts!r}, '
                f'bonder_smarts={self.bonder_smarts!r}, '
                f'del_smarts={self.del_smarts!r})')

    def __str__(self):
        return f'FGInfo({self.name!r})'


class FunctionalGroup:
    """
    Represents the functional group of a molecule.

    Attributes
    ----------
    id : :class:`int`
        The id of the functional group.

    atom_ids : :class:`tuple` of :class:`int`
        The ids of atoms in the functional group.

    bonder_ids : :class:`tuple` of :class:`int`
        The ids of bonder atoms in the functional group.

    deleter_ids : :class:`tuple` of :class:`int`
        The ids of deleter atoms in the functional group.

    info : :class:`FGInfo`
        The :class:`FGInfo` of the functional group type.

    """

    def __init__(self, id_, atom_ids, bonder_ids, deleter_ids, info):
        """
        Initialize a functional group.

        Parameters
        ----------
        id_ : :class:`int`
            The id of the functional group.

        atom_ids : :class:`tuple` of :class:`int`
            The ids of atoms in the functional group.

        bonder_ids : :class:`tuple` of :class:`int`
            The ids of bonder atoms in the functional group.

        deleter_ids : :class:`tuple` of :class:`int`
            The ids of deleter atoms in the functional group.

        info : :class:`FGInfo` or :class:`str`
            The :class:`FGInfo` of the functional group to which the
            functional group belongs. Can also be the name of the
            :class:`FGInfo`.

        """

        self.id = id_
        self.atom_ids = atom_ids
        self.bonder_ids = bonder_ids
        self.deleter_ids = deleter_ids

        if isinstance(info, str):
            self.info = next(fg_info for fg_info in functional_groups
                             if fg_info.name == info)
        else:
            self.info = info

    def shifted_fg(self, id_, shift):
        """
        Create a new :class:`FunctionalGroup` with shifted ids.

        Parameters
        ----------
        id_ : :class:`int`
            The id of the new functional group.

        shift : :class:`int`
            The number to shift the atom ids by.

        Returns
        -------
        :class:`FunctionalGroup`
            A :class:`FunctionalGroup` with all the atom ids atoms
            shifted upward by `shift`.

        """

        atom_ids = tuple(id + shift for id in self.atom_ids)
        bonder_ids = tuple(id + shift for id in self.bonder_ids)
        deleter_ids = tuple(id + shift for id in self.deleter_ids)

        return self.__class__(id_=id_,
                              atom_ids=atom_ids,
                              bonder_ids=bonder_ids,
                              deleter_ids=deleter_ids,
                              info=self.info)

    def remove_deleters(self, deleters):
        """
        Update and remove atom ids based on `deleters`.

        For each deleter atom that is smaller than an atom id, the
        atom id is decreased by 1. If the atom id is equal to a
        `deleters` atom id, it is removed.

        Parameters
        ----------
        deleters : :class:`list` of :class:`int`
            Ids of atoms which are being removed from the molecule.
            Must be sorted in ascending order.

        Returns
        -------
        None : :class:`NoneType`

        """

        self.atom_ids = self._remove_deleters(self.atom_ids,
                                              deleters)

        self.bonder_ids = self._remove_deleters(self.bonder_ids,
                                                deleters)

        self.deleter_ids = self._remove_deleters(self.deleter_ids,
                                                 deleters)

    def _remove_deleters(self, atom_ids, deleters):
        """
        Update and remove atom ids based on `deleters`.

        For each deleter atom that is smaller than an atom id, the
        atom id is decreased by 1. If the atom id is equal to a
        `deleters` atom id, it is removed.

        Parameters
        ----------
        atom_ids : :class:`tuple` of :class:`int`
            The atom ids which need to be updated.

        deleters : :class:`list` of :class:`int`
            The ids of atoms which are being removed. Must be sorted in
            ascending order.

        Returns
        -------
        :class:`tuple` of :class:`int`
            The atom ids in `atom_ids` after the update.

        """

        # Map each atom id to the number of smaller deleter ids.
        # If the atom id is to be deleted, map to None.
        id_changes = {atom_id: 0 for atom_id in atom_ids}

        for atom_id in atom_ids:
            for deleter in deleters:
                if atom_id > deleter:
                    id_changes[atom_id] += 1
                elif atom_id == deleter:
                    id_changes[atom_id] = None
                # Because the deleters are sorted, there will be no
                # more deleters which are smaller than atom_id.
                if atom_id < deleter:
                    break

        # Apply the id changes to the original atom ids.
        new_atom_ids = []
        for atom_id in atom_ids:
            change = id_changes[atom_id]
            if change is None:
                continue
            else:
                new_atom_ids.append(atom_id - change)

        return tuple(new_atom_ids)

    def __eq__(self, other):
        return (self.id == other.id and
                self.atom_ids == other.atom_ids and
                self.bonder_ids == other.bonder_ids and
                self.deleter_ids == other.deleter_ids and
                self.info == other.info)

    def __hash__(self):
        return id(self)

    def __repr__(self):
        return (f"FunctionalGroup(id_={self.id!r}, "
                f"atom_ids={self.atom_ids!r}, "
                f"bonder_ids={self.bonder_ids!r}, "
                f"deleter_ids={self.deleter_ids!r}, "
                f"info={self.info.name!r})")

    def __str__(self):
        return repr(self)


class Reactor:
    """
    Performs reactions between functional groups of a molecule.

    This class is responsible for reacting the functional groups in a
    molecule during assembly. First, an instance of this class is
    initialized with a :class:`rdkit.Mol`. This is the molecule which
    is going to have atom and bonds added and removed.

    .. code-block:: python

        mol = rdkit.MolFromMolFile(...)
        reactor = Reactor(mol)

    We force the molecule to have atoms and bonds add or removed
    between certain functional groups by using the :meth:`react`
    method.

    .. code-block:: python

        # Represents a functional group found in mol.
        fg1 = FunctionalGroup(id_=0,
                              atom_ids=[1, 34, 3],
                              bonder_ids=[1],
                              deleter_ids=[34, 3],
                              info=FGInfo('amine')
        )

        # Represents another functional group found in mol.
        fg2 = FunctionalGroup(id_=1,
                              atom_ids=[10, 2, 12],
                              bonder_ids=[12],
                              deleter_ids=[2, 10],
                              info=FGInfo('aldehyde')
        )

        # Carry out a reaction between the atoms in fg1 and fg2.
        reactor.react(fg1, fg2)

        # Lets assume there a further functional groups in mol that
        # we wish to react.
        fg3 = FunctionalGroup(...)
        fg4 = FunctionalGroup(...)
        fg5 = FunctionalGroup(...)
        fg6 = FunctionalGroup(...)
        fg7 = FunctionalGroup(...)

        reactor.react(fg3, fg4)

        # Some reactions can take multiple functional groups.
        # You can put in as many functional groups as you like, given
        # an appropriate reaction is defined.
        reactor.react(fg5, fg6, fg7)

    Once we are done carrying out reactions on the molecule we can
    get the resulting molecule.

    .. code-block:: python

        # product is an rdkit molecule, with the earlier reactions
        # carried out and all deleter atoms removed.
        product = reactor.result(del_atoms=True)

    Note that after :meth:`result` is run, all :class:`FunctionalGroup`
    instances which were passed to :meth:`react` have their atom ids
    updated to account for any atoms which have been deleted. If
    there are functional groups which were not reacted, but have
    atom ids which should be updated, they should be added to
    :attr:`func_groups`.

    An obvious question given this tutorial, is what reaction does
    :meth:`react` carry out? This is documented by :meth:`react`.
    However, react in most cases, will carry out a default reaction,
    which adds a bond between the bonder atoms of two functional
    groups. The bond order of the added bond is single by default but
    can be modified by editing :attr:`bond_orders`. Here you will
    specify the :class:`ReactionKey` for a reaction and what bond order
    you want that reaction to use.

    For some reactions you may wish to forgo the default reaction and
    do something more complex. This is neccessary because some
    reactions cannot be described by the simple combination of adding a
    bond while deleting some existing atoms. For example,
    consider the aldol reaction:

        CH3C(=O)CH3 + CH3C(=O)CH3 --> CH3(=O)CH2C(OH)(CH3)CH3

    Here a ketone is converted into an alcohol. If you wish to
    support a complex reaction, add it as a method within this class.
    The method will need take some :class:`FunctionalGroup` instances
    as arguments. These are the functional groups which react. Within
    the method itself, the attribute :attr:`bonds_made` must be
    incremented by the number of bonds added by the reaction. Beyond
    that, the method should operate :attr:`emol` and add and remove
    bonds from it as necessary. If the method adds atoms to
    :attr:`emol` it should update :attr:`new_atom_coords`.

    Once the method is defined, :attr:`custom_reactions` needs to
    be updated.

    Attributes
    ----------
    bond_orders : :class:`dict`
        When the default reaction is performed by :meth:`react`,
        if the bond added between the two functional groups is not
        single, the desired bond order should be placed in this
        dictionary. The dictionary maps the reaction's
        :class:`ReactionKey` to the desired bond order.

    custom_reactions : :class:`dict`
        Maps a :class:`ReactionKey` for a given reaction to a custom
        method which carries out the reaction. This means that
        :meth:`react` will use that method for carrying out that
        reaction instead.

    periodic_custom_reactions : :class:`dict`
        Maps a :class:`ReactionKey` for a given reaction to a custom
        method which carries out the reaction. This means that
        :meth:`periodic_react` will use that method for carrying out
        that reaction instead.

    mol : :class:`rdkit.Mol`
        The molecule on which the reactor adds and removes atoms and
        bonds.

    emol : :class:`rdkit.EditableMol`
        An editable version of :attr:`mol`. Used for adding and
        removing atoms and bonds.

    periodic_bonds : :class:`list` of :class:`.AtomicPeriodicBond`
        The periodic bonds added by the reactor.

    bonds_made : :class:`int`
        The number of bonds added.

    new_atom_coords : :class:`list` of :class:`tuple`
        When a new atom is added by the reactor, its desired
        desired coordinates are placed here. The :class:`list`
        has the form

        .. code-block:: python

            new_atom_coords = [
                (32, np.array([12.1, 3.5, 0.1])),
                (41, np.array([13.1, -42.1, 2.]))
            ]

        where the first element of each tuple is the atom id of a
        newly added atom and the second element is a
        :class:`numpy.ndarray` holding its desired coordinate.

    deleters : :class:`list` of :class:`int`
        The ids of atoms which are to be removed.

    func_groups : :class:`list` of :class:`FunctionalGroup`
        The functional groups which the reactor keeps up to date.

    """

    double = rdkit.rdchem.BondType.DOUBLE
    triple = rdkit.rdchem.BondType.TRIPLE
    bond_orders = {
        ReactionKey('amine', 'aldehyde'): double,
        ReactionKey('amide', 'aldehyde'): double,
        ReactionKey('nitrile', 'aldehyde'): double,
        ReactionKey('amide', 'amine'): double,
        ReactionKey('terminal_alkene', 'terminal_alkene'): double,
        ReactionKey('alkyne2', 'alkyne2'): triple
    }

    def __init__(self, mol=None):
        """
        Initialize a :class:`Reactor`.

        Parameters
        ----------
        mol : :class:`rdkit.Mol`, optional
            The molecule on which the reactor adds and removes atoms
            and bonds.

        """

        self.custom_reactions = {

            ReactionKey('boronic_acid', 'diol'):
                self.boronic_acid_with_diol,

            ReactionKey('diol', 'difluorene'):
                partial(self.diol_with_dihalogen,
                        dihalogen='difluorene'),

            ReactionKey('diol', 'dibromine'):
                partial(self.diol_with_dihalogen,
                        dihalogen='dibromine'),

            ReactionKey('ring_amine', 'ring_amine'):
                self.ring_amine_with_ring_amine

        }

        self.periodic_custom_reactions = {}

        if mol is not None:
            self.set_molecule(mol)

        self.periodic_bonds = []
        self.bonds_made = 0
        self.new_atom_coords = []
        self.deleters = []
        self.func_groups = []

    def set_molecule(self, mol):
        """
        Update :attr:`mol` and :attr:`emol`.

        Parameters
        ----------
        mol : :class:`rdkit.Mol`, optional
            The molecule on which the reactor adds and removes atoms
            and bonds.

        Returns
        -------
        None : :class:`NoneType`

        """

        self.mol = mol
        self.emol = rdkit.EditableMol(mol)

    def react(self, *fgs, track_fgs=True):
        """
        Creates bonds between functional groups.

        This function first looks at the functional groups provided via
        the `*fgs` argument and checks which functional groups are
        involved in the reaction. If the functional groups are handled
        by one of the custom reactions specified in
        :attr:`custom_reactions` then that function is executed.

        In all other cases the function is assumed to have received two
        functional groups to react. In these functional groups, the
        bonder atoms have a bond added. The bond is single,
        unless otherwise specified in :attr:`bond_orders`.

        Parameters
        ----------
        *fgs : :class:`FunctionalGroup`
            The functional groups to react.

        track_fgs : :class:`bool`, optional
            Toggles if functional groups are added to
            :attr:`func_groups`.

        Returns
        -------
        None : :class:`NoneType`

        """

        self.deleters.extend(flatten(fg.deleter_ids for fg in fgs))
        if track_fgs:
            self.func_groups.extend(fgs)

        names = (fg.info.name for fg in fgs)
        reaction_key = ReactionKey(*names)
        if reaction_key in self.custom_reactions:
            return self.custom_reactions[reaction_key](*fgs)

        bond = self.bond_orders.get(reaction_key,
                                    rdkit.rdchem.BondType.SINGLE)
        fg1, fg2 = fgs
        self.emol.AddBond(fg1.bonder_ids[0], fg2.bonder_ids[0], bond)
        self.bonds_made += 1

    def periodic_react(self, direction, *fgs, track_fgs=True):
        """
        Like :func:`react` but returns periodic bonds.

        As periodic bonds are returned, no bonds are added to
        :attr:`emol`.

        Parameters
        ----------
        direction : :class:`list` of :class:`int`
            A 3 member list describing the axes along which the created
            bonds are periodic. For example, ``[1, 0, 0]`` means that
            the bonds are periodic along the x axis in the positive
            direction.

        *fgs : :class:`FunctionalGroup`
            The functional groups to react.

        track_fgs : :class:`bool`, optional
            Toggles if functional groups are added to
            :attr:`func_groups`.

        Returns
        -------
        None : :class:`NoneType`

        """

        self.deleters.extend(flatten(fg.deleter_ids for fg in fgs))
        if track_fgs:
            self.func_groups.extend(fgs)

        names = (fg.info.name for fg in fgs)
        reaction_key = ReactionKey(*names)
        if reaction_key in self.periodic_custom_reactions:
            rxn_fn = self.periodic_custom_reactions[reaction_key]
            return rxn_fn(direction, *fgs)

        bond = self.bond_orders.get(reaction_key,
                                    rdkit.rdchem.BondType.SINGLE)

        # Make sure the direction of the periodic bond is maintained.
        fg1, fg2 = fgs
        self.periodic_bonds.append(
            AtomicPeriodicBond(fg1.bonder_ids[0],
                               fg2.bonder_ids[0],
                               bond,
                               direction)
        )
        self.bonds_made += 1

    def result(self, del_atoms):
        """
        Creates the molecule after all reactions have been done.

        This method will also update all the functional groups
        passed with :meth:`react` calls. It will update the atom
        ids to account for the fact that atoms have been deleted.

        Parameters
        ----------
        del_atoms : :class:`bool`
            Toggles if deleter atoms should be removed from the
            product molecule.

        Returns
        -------
        :class:`rdkit.Mol`
            The product molecule.

        """

        # If new atoms were added, update the positions in the
        # conformer.
        if self.new_atom_coords:
            self.mol = self.emol.GetMol()
            conf = self.mol.GetConformer()
            for atom_id, coord in self.new_atom_coords:
                point3d = rdkit_geo.Point3D(*coord)
                conf.SetAtomPosition(atom_id, point3d)
            self.emol = rdkit.EditableMol(self.mol)

        if del_atoms:
            # Needs to be sorted for fg.remove_deleters.
            deleters = sorted(self.deleters)

            # Go in reverse order else atom ids change during loop.
            for atom_id in reversed(deleters):
                self.emol.RemoveAtom(atom_id)

            # Update all the functional groups to account for the fact
            # that atoms have been removed.
            for fg in self.func_groups:
                fg.remove_deleters(deleters)

        return self.emol.GetMol()

    def diol_with_dihalogen(self, fg1, fg2, dihalogen):
        """
        Creates bonds between functional groups.

        Parameters
        ----------
        fg1 : :class:`.FunctionalGroup`
            A functional group which undergoes the reaction.

        fg2 : :class:`.FunctionalGroup`
            A functional group which undergoes the reaction.

        halogen : :class:`str`
            The name of the dihalogen functional group to use. For example,
            ``'dibromine'`` or ``'difluorene'``.

        Returns
        -------
        None : :class:`NoneType`

        """

        conf = self.mol.GetConformer()
        bond = rdkit.rdchem.BondType.SINGLE

        diol = fg1 if fg1.info.name == 'diol' else fg2
        dihalogen = fg2 if diol is fg1 else fg1

        distances = []
        for carbon in dihalogen.bonder_ids:
            c_coord = np.array([*conf.GetAtomPosition(carbon)])
            for oxygen in diol.bonder_ids:
                o_coord = np.array([*conf.GetAtomPosition(oxygen)])
                d = euclidean(c_coord, o_coord)
                distances.append((d, carbon, oxygen))
        distances.sort()

        deduped_pairs = []
        seen_o, seen_c = set(), set()
        for d, c, o in distances:
            if c not in seen_c and o not in seen_o:
                deduped_pairs.append((c, o))
                seen_c.add(c)
                seen_o.add(o)

        (c1, o1), (c2, o2), *_ = deduped_pairs
        assert c1 != c2 and o1 != o2
        self.emol.AddBond(c1, o1, bond)
        self.emol.AddBond(c2, o2, bond)
        self.bonds_made += 2

    def boronic_acid_with_diol(self, fg1, fg2):
        """
        Creates bonds between functional groups.

        Parameters
        ----------
        fg1 : :class:`.FunctionalGroup`
            A functional group which undergoes the reaction.

        fg2 : :class:`.FunctionalGroup`
            A functional group which undergoes the reaction.

        Returns
        -------
        None : :class:`NoneType`

        """

        bond = rdkit.rdchem.BondType.SINGLE

        boron = fg1 if fg1.info.name == 'boronic_acid' else fg2
        diol = fg2 if boron is fg1 else fg1

        boron_atom = boron.bonder_ids[0]
        self.emol.AddBond(boron_atom, diol.bonder_ids[0], bond)
        self.emol.AddBond(boron_atom, diol.bonder_ids[1], bond)
        self.bonds_made += 2

    def ring_amine_with_ring_amine(self, fg1, fg2):
        """
        Creates bonds between functional groups.

        Parameters
        ----------
        fg1 : :class:`.FunctionalGroup`
            A functional group which undergoes the reaction.

        fg2 : :class:`.FunctionalGroup`
            A functional group which undergoes the reaction.

        Returns
        -------
        None : :class:`NoneType`

        """

        c1 = next(a for a in fg1.bonder_ids
                  if self.mol.GetAtomWithIdx(a).GetAtomicNum() == 6)
        n1 = next(a for a in fg1.bonder_ids
                  if self.mol.GetAtomWithIdx(a).GetAtomicNum() == 7)

        c2 = next(a for a in fg2.bonder_ids
                  if self.mol.GetAtomWithIdx(a).GetAtomicNum() == 6)
        n2 = next(a for a in fg2.bonder_ids
                  if self.mol.GetAtomWithIdx(a).GetAtomicNum() == 7)

        conf = self.mol.GetConformer()
        n1_pos = np.array([*conf.GetAtomPosition(n1)])
        n2_pos = np.array([*conf.GetAtomPosition(n2)])

        c1_pos = np.array([*conf.GetAtomPosition(c1)])
        c2_pos = np.array([*conf.GetAtomPosition(c2)])

        n_joiner = self.emol.AddAtom(rdkit.Atom(6))
        n_joiner_pos = (n1_pos + n2_pos) / 2
        self.new_atom_coords.append((n_joiner, n_joiner_pos))

        nh1 = self.emol.AddAtom(rdkit.Atom(1))
        nh1_pos = n_joiner_pos + np.array([0, 0, 1])
        self.new_atom_coords.append((nh1, nh1_pos))

        nh2 = self.emol.AddAtom(rdkit.Atom(1))
        nh2_pos = n_joiner_pos + np.array([0, 0, -1])
        self.new_atom_coords.append((nh2, nh2_pos))

        nc_joiner1 = self.emol.AddAtom(rdkit.Atom(6))
        nc_joiner1_pos = (c1_pos + n2_pos) / 2
        self.new_atom_coords.append((nc_joiner1, nc_joiner1_pos))

        nc1h1 = self.emol.AddAtom(rdkit.Atom(1))
        nc1h1_pos = nc_joiner1_pos + np.array([0, 0, 1])
        self.new_atom_coords.append((nc1h1, nc1h1_pos))

        nc1h2 = self.emol.AddAtom(rdkit.Atom(1))
        nc1h2_pos = nc_joiner1_pos + np.array([0, 0, -1])
        self.new_atom_coords.append((nc1h2, nc1h2_pos))

        nc_joiner2 = self.emol.AddAtom(rdkit.Atom(6))
        nc_joiner2_pos = (c2_pos + n1_pos) / 2
        self.new_atom_coords.append((nc_joiner2, nc_joiner2_pos))

        nc2h1 = self.emol.AddAtom(rdkit.Atom(1))
        nc2h1_pos = nc_joiner2_pos + np.array([0, 0, 1])
        self.new_atom_coords.append((nc2h1, nc2h1_pos))

        nc2h2 = self.emol.AddAtom(rdkit.Atom(1))
        nc2h2_pos = nc_joiner2_pos + np.array([0, 0, -1])
        self.new_atom_coords.append((nc2h2, nc2h2_pos))

        single = rdkit.rdchem.BondType.SINGLE
        self.emol.AddBond(n1, n_joiner, single)
        self.emol.AddBond(n2, n_joiner, single)
        self.emol.AddBond(n_joiner, nh1, single)
        self.emol.AddBond(n_joiner, nh2, single)

        self.emol.AddBond(c1, nc_joiner1, single)
        self.emol.AddBond(n2, nc_joiner1, single)
        self.emol.AddBond(nc_joiner1, nc1h1, single)
        self.emol.AddBond(nc_joiner1, nc1h2, single)

        self.emol.AddBond(c2, nc_joiner2, single)
        self.emol.AddBond(n1, nc_joiner2, single)
        self.emol.AddBond(nc_joiner2, nc2h1, single)
        self.emol.AddBond(nc_joiner2, nc2h2, single)

        self.bonds_made += 6


functional_groups = (

    FGInfo(name="amine",
           fg_smarts="[N]([H])[H]",
           bonder_smarts=[Match(smarts="[$([N]([H])[H])]", n=1)],
           del_smarts=[Match(smarts="[$([H][N][H])]", n=2)]),

    FGInfo(name="aldehyde",
           fg_smarts="[C](=[O])[H]",
           bonder_smarts=[Match(smarts="[$([C](=[O])[H])]", n=1)],
           del_smarts=[Match(smarts="[$([O]=[C][H])]", n=1)]),

    FGInfo(name="carboxylic_acid",
           fg_smarts="[C](=[O])[O][H]",
           bonder_smarts=[Match(smarts="[$([C](=[O])[O][H])]", n=1)],
           del_smarts=[Match(smarts="[$([H][O][C](=[O]))]", n=1),
                       Match(smarts="[$([O]([H])[C](=[O]))]", n=1)]),


    FGInfo(name="amide",
           fg_smarts="[C](=[O])[N]([H])[H]",
           bonder_smarts=[
                Match(smarts="[$([C](=[O])[N]([H])[H])]", n=1)
           ],
           del_smarts=[
                Match(smarts="[$([N]([H])([H])[C](=[O]))]", n=1),
                Match(smarts="[$([H][N]([H])[C](=[O]))]", n=2)
           ]),

    FGInfo(name="thioacid",
           fg_smarts="[C](=[O])[S][H]",
           bonder_smarts=[Match(smarts="[$([C](=[O])[S][H])]", n=1)],
           del_smarts=[Match(smarts="[$([H][S][C](=[O]))]", n=1),
                       Match(smarts="[$([S]([H])[C](=[O]))]", n=1)]),

    FGInfo(name="alcohol",
           fg_smarts="[O][H]",
           bonder_smarts=[Match(smarts="[$([O][H])]", n=1)],
           del_smarts=[Match(smarts="[$([H][O])]", n=1)]),

    FGInfo(name="thiol",
           fg_smarts="[S][H]",
           bonder_smarts=[Match(smarts="[$([S][H])]", n=1)],
           del_smarts=[Match(smarts="[$([H][S])]", n=1)]),

    FGInfo(name="bromine",
           fg_smarts="*[Br]",
           bonder_smarts=[Match(smarts="[$(*[Br])]", n=1)],
           del_smarts=[Match(smarts="[$([Br]*)]", n=1)]),

    FGInfo(name="iodine",
           fg_smarts="*[I]",
           bonder_smarts=[Match(smarts="[$(*[I])]", n=1)],
           del_smarts=[Match(smarts="[$([I]*)]", n=1)]),

    FGInfo(name='alkyne',
           fg_smarts='[C]#[C][H]',
           bonder_smarts=[Match(smarts='[$([C]([H])#[C])]', n=1)],
           del_smarts=[Match(smarts='[$([H][C]#[C])]', n=1)]),

    FGInfo(name='terminal_alkene',
           fg_smarts='[C]=[C]([H])[H]',
           bonder_smarts=[Match(smarts='[$([C]=[C]([H])[H])]', n=1)],
           del_smarts=[Match(smarts='[$([H][C]([H])=[C])]', n=2),
                       Match(smarts='[$([C](=[C])([H])[H])]', n=1)]),

    FGInfo(name='boronic_acid',
           fg_smarts='[B]([O][H])[O][H]',
           bonder_smarts=[Match(smarts='[$([B]([O][H])[O][H])]', n=1)],
           del_smarts=[Match(smarts='[$([O]([H])[B][O][H])]', n=2),
                       Match(smarts='[$([H][O][B][O][H])]', n=2)]),

    # This amine functional group only deletes one of the
    # hydrogen atoms when a bond is formed.
    FGInfo(name="amine2",
           fg_smarts="[N]([H])[H]",
           bonder_smarts=[Match(smarts="[$([N]([H])[H])]", n=1)],
           del_smarts=[Match(smarts="[$([H][N][H])]", n=1)]),

    FGInfo(name="secondary_amine",
           fg_smarts="[H][N]([#6])[#6]",
           bonder_smarts=[
                Match(smarts="[$([N]([H])([#6])[#6])]", n=1)
           ],
           del_smarts=[Match(smarts="[$([H][N]([#6])[#6])]", n=1)]),

    FGInfo(name='diol',
           fg_smarts='[H][O][#6]~[#6][O][H]',
           bonder_smarts=[
               Match(smarts='[$([O]([H])[#6]~[#6][O][H])]', n=2)
           ],
           del_smarts=[
               Match(smarts='[$([H][O][#6]~[#6][O][H])]', n=2)
           ]),

    FGInfo(name='difluorene',
           fg_smarts='[F][#6]~[#6][F]',
           bonder_smarts=[Match(smarts='[$([#6]([F])~[#6][F])]', n=2)],
           del_smarts=[Match(smarts='[$([F][#6]~[#6][F])]', n=2)]),

    FGInfo(name='dibromine',
           fg_smarts='[Br][#6]~[#6][Br]',
           bonder_smarts=[
               Match(smarts='[$([#6]([Br])~[#6][Br])]', n=2)
           ],
           del_smarts=[Match(smarts='[$([Br][#6]~[#6][Br])]', n=2)]),

    FGInfo(name='alkyne2',
           fg_smarts='[C]#[C][H]',
           bonder_smarts=[Match(smarts='[$([C]#[C][H])]', n=1)],
           del_smarts=[Match(smarts='[$([H][C]#[C])]', n=1),
                       Match(smarts='[$([C](#[C])[H])]', n=1)]),

    FGInfo(name='ring_amine',
           fg_smarts='[N]([H])([H])[#6]~[#6]([H])~[#6R1]',
           bonder_smarts=[
               Match(
                   smarts='[$([N]([H])([H])[#6]~[#6]([H])~[#6R1])]',
                   n=1
               ),
               Match(
                   smarts='[$([#6]([H])(~[#6R1])~[#6][N]([H])[H])]',
                   n=1
               ),
            ],
           del_smarts=[
               Match(
                   smarts='[$([H][N]([H])[#6]~[#6]([H])~[#6R1])]',
                   n=2
               ),
               Match(
                   smarts='[$([H][#6](~[#6R1])~[#6][N]([H])[H])]',
                   n=1
               )
           ])

    )

functional_group_infos = {fg.name: fg for fg in functional_groups}



from .energy_calculators import EnergyCalculator


class MOPACEnergy(EnergyCalculator):
    def __init__(self, mopac_path):
        self.mopac_path = mopac_path

    def energy(self, mol, conformer=-1):
        """
        Calculates the energy using MOPAC.

        Note that this requires MOPAC to be installed and have a
        valid license.

        Parameters
        ----------
        settings : :class:`dict`, optional
            A dictionary which maps the names of the optimization
            parameters to their values. Valid values are:

                'hamiltonian' : :class:`str` (default = ``'PM7'``
                    A series of different methods can be selected:
                    PM7, PM6, AM1, CIS (CISD, CISDT), MNDO, RM1, etc..

                    PM7 is the latest version of the reparametrization
                    of NDDO theory, where all the atomic and diatomic
                    parameters were re-optimized / updated from PM6
                    [#]_.

                'eps' : :class:`float` (default = ``80.1``)
                    Sets the dielectric constant for the solvent.
                    Presence of this keyword will cause the COSMO
                    (Conductor-like Screening Model) method to be used
                    to approximate the effect of a solvent model
                    surrounding the molecule. Solvents with a low
                    dielectric constant are not likely to work well
                    with this model. ``0`` means that the dielectric
                    constant is not included in the calculation.
                    ``80.1`` can be used to model a water environment
                    at room temperature.

                'charge' : :class:`float` (default = ``0``)
                    The charge of the system.

                'timeout' : :class:`float` (default = ``172800``)
                    The amount in seconds the calculation is allowed to
                    run before being terminated. The default value is
                    ``2`` days or ``172,800`` seconds.

        mopac_path : :class:`str`
            The full path to the MOPAC installation.

        Returns
        -------
        :class:`float`
            The calculated energy.

        References
        ----------
        .. [#] http://openmopac.net/PM7_accuracy/PM7_accuracy.html

        """

        if settings is None:
            settings = {}

        # Define default vals for the MOPAC input
        vals = {
                'hamiltonian': 'PM7',
                'method': 'NOOPT',
                'eps': 80.1,
                'charge': 0,
                'timeout': 172800,
                }

        vals.update(settings)

        # To prevent conflicts when running this function in parallel,
        # a temporary copy of the molecular structure file is made and
        # used for mopac calculations.

        # Unique file name is generated by inserting a random int into
        # the file path.
        tmp_file = "{}.mol".format(uuid4().int)
        self.molecule.write(tmp_file)

        file_root, ext = os.path.splitext(tmp_file)

        # Generate the input file
        _create_mop(file_root, self.molecule, vals)
        # Run MOPAC
        _run_mopac(file_root, mopac_path)
        return _extract_MOPAC_en(file_root)



"""

"""

from .base import Topology
from ...utilities import translation_component, quaternion
import numpy as np
import rdkit.Chem.AllChem as rdkit


class Dimer(Topology):
    def __init__(self, u, distance):
        """

        """

        self.u = u
        self.distance = distance

    def place_mols(self, macro_mol):
        bb = macro_mol.building_blocks[0]
        bb.set_position([0, 0, 0])
        macro_mol.mol = rdkit.Mol(bb.mol)

        # The axis of rotation shoudl be the longest dimension of the
        # building block.
        _, a1, a2 = bb.max_diameter()
        a1_coord = bb.atom_coords(int(a1))
        a2_coord = bb.atom_coords(int(a2))
        axis = a1_coord - a2_coord
        bb.rotate(np.random.random()*np.pi/2, axis)
        q = quaternion(self.u)
        v = translation_component(q)*self.distance
        new_pos = (bb.mol.GetConformer().GetPositions() +
                   np.matrix(v)).T
        bb.set_position_from_matrix(new_pos)
        macro_mol.mol = rdkit.CombineMols(macro_mol.mol, bb.mol)

    def bonded_fgs(self, macro_mol):
        return []



"""
Defines MacroModel optimizers.

"""

import os
import subprocess as sp
import time
import rdkit.Chem.AllChem as rdkit
import psutil
import re
from uuid import uuid4
import logging
import gzip

from ...utilities import MAEExtractor, move_generated_macromodel_files
from .optimizers import Optimizer

logger = logging.getLogger(__name__)


class MacroModelConversionError(Exception):
    ...


class MacroModelPathError(Exception):
    ...


class MacroModelForceFieldError(Exception):
    ...


class MacroModelOptimizationError(Exception):
    ...


class MacroModelLewisStructureError(Exception):
    ...


class _MacroModel(Optimizer):
    """
    Base class for MacroModel optimzers.

    Attributes
    ----------
    macromodel_path : :class:`str`
        The full path of the Schrodinger suite within the user's
        machine. For example, on a Linux machine this may be something
        like ``'/opt/schrodinger2017-2'``.

    output_dir : :class:`str`
        The name of the directory into which files generated during
        the optimization are written, if ``None`` then
        :func:`uuid.uuid4` is used.

    force_field : :class:`int`
        The number of the force field to be used.

    maximum_iterations : :class:`int`
        The maximum number of iterations done during the optimization.

    minimum_gradient : :class:`float`
        The gradient at which optimization is stopped.

    restricted : :class:`bool`
        If ``False`` then all bonds are optimized, not just the ones
        created during macromolecular assembly. If ``True`` then an
        optimization is performed only on the bonds added during
        molecular assembly. If ``'both'`` then a restricted
        optimization is performed first, followed by a regular
        optimization.

    timeout : :class:`float`
        The amount in seconds the MD is allowed to run before
        being terminated. ``None`` means there is no timeout.

    """

    def __init__(self,
                 macromodel_path,
                 output_dir,
                 restricted,
                 timeout,
                 force_field,
                 maximum_iterations,
                 minimum_gradient,
                 use_cache):
        """
        Initializes a :class:`_MacroModel` instance.

        Parameters
        ----------
        macromodel_path : :class:`str`
            The full path of the Schrodinger suite within the user's
            machine. For example, on a Linux machine this may be
            something like ``'/opt/schrodinger2017-2'``.

        output_dir : :class:`str`, optional
            The name of the directory into which files generated during
            the optimization are written, if ``None`` then
            :func:`uuid.uuid4` is used.

        restricted : :class:`bool`, optional
            If ``False`` then all bonds are optimized, not just the
            ones created during macromolecular assembly. If ``True``
            then an optimization is performed only on the bonds added
            during molecular assembly. If ``'both'`` then a restricted
            optimization is performed first, followed by a regular
            optimization.

        timeout : :class:`float`, optional
            The amount in seconds the optimization is allowed to run
            before being terminated. ``None`` means there is no
            timeout.

        force_field : :class:`int`, optional
            The number of the force field to be used.

        maximum_iterations : :class:`int`, optional
            The maximum number of iterations done during the
            optimization.

        minimum_gradient : :class:`float`, optional
            The gradient at which optimization is stopped.

        use_cache : :class:`bool`, optional
            If ``True`` :meth:`optimize` will not run twice on the same
            molecule and conformer.

        """

        self.macromodel_path = macromodel_path
        self.output_dir = output_dir
        self.restricted = restricted
        self.timeout = timeout
        self.force_field = force_field
        self.maximum_iterations = maximum_iterations
        self.minimum_gradient = minimum_gradient
        super().__init__(use_cache=use_cache)

    def run_bmin(self, mol):
        """
        Runs an optimization using bmin.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule being optimized.

        Returns
        -------
        None : :class:`NoneType`

        Raises
        ------
        :class:`MacroModelOptimizationError`
            If the optimization failed for some unspecified.

        :class:`MacroModelForceFieldError`
            If the force field could not be used with the molecule.

        :class:`MacroModelLewisStructureError`
            If Lewis structure of the molecule had issues.

        :class:`MacroModelPathError`
            If an invalid MacroModel path is being used.

        """

        logger.info(f'Running bmin on "{mol.name}".')

        # To run MacroModel a command is issued to the console via
        # ``subprocess.Popen``. The command is the full path of the
        # ``bmin`` program. ``bmin`` is located in the Schrodinger
        # installation folder.
        file_root, ext = os.path.splitext(mol._file)
        log_file = f'{file_root}.log'
        opt_app = os.path.join(self.macromodel_path, 'bmin')
        # The first member of the list is the command, the following
        # ones are any additional arguments.

        opt_cmd = [opt_app, file_root, '-WAIT', '-LOCAL']

        incomplete = True
        while incomplete:
            process = psutil.Popen(opt_cmd,
                                   stdout=sp.PIPE,
                                   stderr=sp.STDOUT,
                                   universal_newlines=True)
            try:
                output, _ = process.communicate(timeout=self.timeout)

            except sp.TimeoutExpired:
                logger.warning(
                    ('Minimization took too long and was terminated '
                     f'by force on "{mol.name}".')
                )
                self.kill_bmin(mol)
                output = ''

            logger.debug(
                f'Output of bmin on "{mol.name}" was: {output}.'
            )

            with open(log_file, 'r') as log:
                log_content = log.read()

            # Check the log for error reports.
            error1 = 'termination due to error condition           21-'
            if error1 in log_content:
                raise MacroModelOptimizationError(
                    'bmin crashed. See log file.'
                )

            error2 = 'FATAL do_nosort_typing: NO MATCH found for atom '
            if error2 in log_content:
                raise MacroModelForceFieldError(
                    'The log implies the force field failed.'
                )

            error3 = (
                'FATAL gen_lewis_structure(): '
                'could not find best Lewis structure'
            )
            error4 = (
                'skipping input structure  '
                'due to forcefield interaction errors'
            )
            if error3 in log_content and error4 in log_content:
                raise MacroModelLewisStructureError(
                        'bmin failed due to poor Lewis structure.'
                )

            if 'MDYN error encountered' in log_content:
                raise MacroModelOptimizationError(
                    'MD error during optimization.'
                )

            # If optimization fails because a wrong Schrodinger path
            # was given, raise.
            if 'The system cannot find the path specified' in output:
                raise MacroModelPathError(
                    'Invalid Schrodinger path given to bmin.'
                )

            # If optimization fails because the license is not found,
            # rerun the function.
            if self.license_found(output, mol):
                incomplete = False

        # Make sure the .maegz file created by the optimization is
        # present.
        maegz = file_root + '-out.maegz'
        self.wait_for_file(maegz)
        if not os.path.exists(log_file) or not os.path.exists(maegz):
            raise MacroModelOptimizationError(
                'The .log and/or .maegz files were not created.'
            )

    def kill_bmin(self, mol):
        """
        Kills bmin.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule being optimized.

        Returns
        -------
        None : :class:`NoneType`

        """

        name, ext = os.path.splitext(mol._file)
        name = re.split(r'\\|/', name)[-1]
        app = os.path.join(self.macromodel_path, 'jobcontrol')
        cmd = [app, '-stop', name]

        incomplete = True
        while incomplete:
            out = sp.run(cmd,
                         stdout=sp.PIPE,
                         stderr=sp.STDOUT,
                         universal_newlines=True)

            # Keep re-running the function until license is found.
            if self.license_found(out.stdout):
                incomplete = False

        # This loop causes the function to wait until the job has been
        # killed via job control. This means the output files will have
        # been written by the time the function exits. Essentially the
        # loop continues until the job is no longer found by
        # "./jobcontrol -list"
        cmd = [app, '-list']
        output = name
        start = time.time()
        while name in output:
            output = sp.run(cmd,
                            stdout=sp.PIPE,
                            stderr=sp.STDOUT,
                            universal_newlines=True).stdout
            if time.time() - start > 600:
                break

    def license_found(self, output, mol=None):
        """
        Checks to see if minimization failed due to a missing license.

        The user can be notified of this in one of two ways. Sometimes
        the output of the submission contains the message informing
        that the license was not found and in other cases it will be
        the log file. This function checks both of these sources for
        this message.

        Parameters
        ----------
        output : :class:`str`
            The output from submitting the minimization of the
            structure to bmin.

        mol : :class:`.Molecule`, optional
            The molecule being optimized. If the ``.log`` file is not
            to be checked, the default ``None`` should be used.

        Returns
        -------
        :class:`bool`
            ``True`` if the license was found. ``False`` if the
            minimization did not occur due to a missing license.

        """

        if 'Could not check out a license for mmlibs' in output:
            return False
        if mol is None:
            return True

        # To check if the log file mentions a missing license file open
        # the log file and scan for the apporpriate string.

        # Check if the file exists first. If not, this is often means
        # the calculation must be redone so return False anyway.
        log_file_path = mol._file.replace('mol', 'log')
        with open(log_file_path, 'r') as log_file:
            log_file = log_file.read()

        if 'Could not check out a license for mmlibs' in log_file:
            return False

        return True

    @staticmethod
    def com_line(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9):
        return (f' {arg1:<5}{arg2:>7}{arg3:>7}'
                f'{arg4:>7}{arg5:>7}{arg6:>11.4f}'
                f'{arg7:>11.4f}{arg8:>11.4f}{arg9:>11.4f}')

    def structconvert(self, iname, oname):
        """
        Uses structconvert to change file type.

        Parameters
        ----------
        iname : :class:`str`
            The name of the input file.

        oname : :class:`str`
            The name of the output file.

        Returns
        -------
        None : :class:`NoneType`

        """

        convrt_app = os.path.join(self.macromodel_path,
                                  'utilities',
                                  'structconvert')
        convrt_cmd = [convrt_app, iname, oname]

        incomplete = True
        while incomplete:

            # Execute the file conversion.
            try:
                convrt_return = sp.run(convrt_cmd,
                                       stdout=sp.PIPE,
                                       stderr=sp.STDOUT,
                                       universal_newlines=True)

            # If conversion fails because a wrong Schrodinger path was
            # given, raise.
            except FileNotFoundError:
                raise MacroModelPathError(
                    'Wrong Schrodinger path supplied to structconvert.'
                )

            if 'File does not exist' in convrt_return.stdout:
                raise MacroModelConversionError(
                    (f'structconvert input file, {iname}, missing. '
                     f'Console output was {convrt_return.stdout}')
                )

            # Keep re-running the function until license is found.
            if self.license_found(convrt_return.stdout):
                incomplete = False

        # If force field failed, raise.
        if 'number 1' in convrt_return.stdout:
            raise MacroModelForceFieldError(convrt_return.stdout)

        self.wait_for_file(oname)
        if not os.path.exists(oname):
            raise MacroModelConversionError(
                (f'Conversion output file {oname} was not found.'
                 f' Console output was {convrt_return.stdout}.')
             )

        return convrt_return

    def create_mae(self, mol):
        """
        Creates the ``.mae`` file holding the molecule to be optimized.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule which is to be optimized. Its molecular
            structure file is converted to a ``.mae`` file. The
            original file is also kept.

        Returns
        -------
        :class:`str`
            The full path of the newly created ``.mae`` file.

        """

        _, ext = os.path.splitext(mol._file)

        logger.debug(f'Converting {ext} of "{mol.name}" to .mae.')

        # Create the name of the new ``.mae`` file. It is the same as
        # the original structure file, including the same path. Only
        # the extensions are different.
        mae_file = mol._file.replace(ext, '.mae')
        self.structconvert(mol._file, mae_file)
        return mae_file

    def wait_for_file(self, filename, timeout=10):
        """
        Stalls until a given file exists or `timeout` expires.

        Parameters
        ----------
        filename : :class:`str`
            The full path of the file which should be waited for.

        timeout : :class:`int` or :class:`float`, optional
            The number of seconds before the function stops waiting and
            returns.

        Returns
        --------
        None : :class:`NoneType`

        """

        t_start = time.time()
        tick = 0
        while True:
            time_taken = time.time() - t_start
            if divmod(time_taken, 5)[0] == tick + 1:
                logger.warning(f'Waiting for "{filename}".')
                tick += 1

            if os.path.exists(filename) or time_taken > timeout:
                break

    def convert_maegz_to_mae(self, mol):
        """
        Converts a ``.maegz`` file to a ``.mae`` file.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule being optimized. The ``.maegz`` file holding
            its optimized structure is converted to a ``.mae`` file.
            Both versions are kept.

        Returns
        -------
        None : :class:`NoneType`

        """

        logger.debug(f'Converting .maegz of "{mol.name}" to .mae.')
        name, ext = os.path.splitext(mol._file)
        # ``out`` is the full path of the optimized ``.mae`` file.
        maegz = name + '-out.maegz'
        # Replace extensions to get the names of the various files.
        mae = name + '.mae'

        gz_file = gzip.open(maegz)
        with open(mae, 'wb') as f:
            f.write(gz_file.read())
        gz_file.close()

    def fix_params(self, mol, com):
        """
        Fix bond distances and angles in ``.com`` file.

        For each bond distance, bond angle and torisional angle that
        does not involve a bond created during assembly a "FX" command
        is added to the body of the ``.com`` file.

        These lines replace the filler line in the main string.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule which is to be optimized.

        com : :class:`str`
            The body of the ``.com`` file which is to have fix commands
            added.

        Returns
        -------
        :class:`str`
            A string holding the body of the ``.com`` file with
            instructions to fix the various bond distances and angles
            as described in the docstring.

        """

        # Make a string to hold all of the ``FX`` lines.
        fix_block = ''

        # If `restricted` is ``False`` do not add a fix block.
        if not self.restricted:
            return com.replace(
                "!!!BLOCK_OF_FIXED_PARAMETERS_COMES_HERE!!!\n",
                fix_block
            )
        # Add lines that fix the bond distance.
        fix_block = self.fix_distances(mol, fix_block)
        # Add lines that fix the bond angles.
        fix_block = self.fix_bond_angles(mol, fix_block)
        # Add lines that fix the torsional angles.
        fix_block = self.fix_torsional_angles(mol, fix_block)

        return com.replace(
            '!!!BLOCK_OF_FIXED_PARAMETERS_COMES_HERE!!!\n',
            fix_block
        )

    def fix_distances(self, mol, fix_block):
        """
        Adds lines fixing bond distances to ``.com`` body.

        Only bond distances which do not involve bonds created during
        assembly are fixed.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        fix_block : :class:`str`
            A string holding fix commands in the ``.com`` file.

        Returns
        -------
        :class:`str`
            A string holding fix commands in the ``.com`` file.

        """

        bonder_ids = set(
            bid for fg in mol.func_groups for bid in fg.bonder_ids
        )

        # Go through all the bonds in the rdkit molecule. If the bond
        # is not between bonder atoms add a fix line to the ``fix_block``.
        # If the bond does invovle two bonder atoms go to the next bond.
        # This is because a bond between 2 bonder atoms was added during
        # assembly and should therefore not be fixed.
        for bond in mol.mol.GetBonds():
            atom1 = bond.GetBeginAtom()
            atom2 = bond.GetEndAtom()

            if (atom1.GetIdx() in bonder_ids and
               atom2.GetIdx() in bonder_ids):
                continue

            # Make sure that the indices are increased by 1 in the .mae
            # file.
            atom1_id = atom1.GetIdx() + 1
            atom2_id = atom2.GetIdx() + 1
            args = ('FXDI', atom1_id, atom2_id, 0, 0, 99999, 0, 0, 0)
            fix_block += self.com_line(*args)
            fix_block += '\n'

        return fix_block

    def fix_bond_angles(self, mol, fix_block):
        """
        Adds lines fixing bond angles to the ``.com`` body.

        All bond angles of the molecule are fixed.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        fix_block : :class:`str`
            A string holding fix commands in the ``.com`` file.

        Returns
        -------
        :class:`str`
            A string holding fix commands in the ``.com`` file.

        """

        # Create a substructure consisting of 3 dummy atoms bonded with
        # 3 dummy bonds. This substructure will match with any 3 atoms
        # which are bonded together with any combination of bonds.
        # These 3 atoms will therefore have a bond angle.
        ba_mol = rdkit.MolFromSmarts('[*]~[*]~[*]')

        # Get the indices of all atoms which have a bond angle.
        # ``ba_atoms`` is a tuple of tuples of the form ((1,2,3),
        # (4,5,6), (7,8,9), ...). Each inner tuple holds the indicies
        # of the atoms which form a bond angle.
        ba_atoms = mol.mol.GetSubstructMatches(ba_mol)

        for atom_ids in ba_atoms:
            atom_ids = [i+1 for i in atom_ids]
            args = ('FXBA', *atom_ids, 99999, 0, 0, 0, 0)
            fix_block += self.com_line(*args)
            fix_block += '\n'

        return fix_block

    def fix_torsional_angles(self, mol, fix_block):
        """
        Adds lines fixing torsional bond angles to the ``.com`` body.

        All torsional angles of the molecule are fixed.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        fix_block : :class:`str`
            A string holding fix commands in the ``.com`` file.

        Returns
        -------
        :class:`str`
            A string holding fix commands in the ``.com`` file.

        """

        # Create a substructure consisting of 4 dummy atoms bonded with
        # 4 dummy bonds. This substructure will match with any 4 atoms
        # which are bonded together with any combination of bonds.
        # These 4 atoms will therefore have a torsinal angle.
        ta_mol = rdkit.MolFromSmarts('[*]~[*]~[*]~[*]')

        # Get the indices of all atoms which have a torsional angle.
        # ``ta_atoms`` as a tuple of tuples of the form ((1,2,3,4),
        # (4,5,6,7), ...). Each inner tuple holds the indicies of the
        # atoms which have a torsional angle.
        ta_atoms = mol.mol.GetSubstructMatches(ta_mol)

        # Apply the fix.
        for atom_ids in ta_atoms:
            atom_ids = [i+1 for i in atom_ids]
            args = ('FXTA', *atom_ids, 99999, 361, 0, 0)
            fix_block += self.com_line(*args)
            fix_block += '\n'

        return fix_block


class MacroModelForceField(_MacroModel):
    """
    Uses MacroModel force fields to optimize molecules.

    """

    def __init__(self,
                 macromodel_path,
                 output_dir=None,
                 restricted='both',
                 timeout=None,
                 force_field=16,
                 maximum_iterations=2500,
                 minimum_gradient=0.05,
                 use_cache=False):
        """
        Initializes a :class:`MacroModelForceField` object.

        Parameters
        ----------
        macromodel_path : :class:`str`
            The full path of the Schrodinger suite within the user's
            machine. For example, on a Linux machine this may be
            something like ``'/opt/schrodinger2017-2'``.

        output_dir : :class:`str`, optional
            The name of the directory into which files generated during
            the optimization are written, if ``None`` then
            :func:`uuid.uuid4` is used.

        restricted : :class:`bool`, optional
            If ``False`` then all bonds are optimized, not just the
            ones created during macromolecular assembly. If ``True``
            then an optimization is performed only on the bonds added
            during molecular assembly. If ``'both'`` then a restricted
            optimization is performed first, followed by a regular
            optimization.

        timeout : :class:`float`, optional
            The amount in seconds the optimization is allowed to run
            before being terminated. ``None`` means there is no
            timeout.

        force_field : :class:`int`, optional
            The number of the force field to be used.

        maximum_iterations : :class:`int`, optional
            The maximum number of iterations done during the
            optimization.

        minimum_gradient : :class:`float`, optional
            The gradient at which optimization is stopped.

        use_cache : :class:`bool`, optional
            If ``True`` :meth:`optimize` will not run twice on the same
            molecule and conformer.

        """

        super().__init__(macromodel_path=macromodel_path,
                         output_dir=output_dir,
                         force_field=force_field,
                         maximum_iterations=maximum_iterations,
                         minimum_gradient=minimum_gradient,
                         restricted=restricted,
                         timeout=timeout,
                         use_cache=use_cache)

    def generate_com(self, mol):
        """
        Create a ``.com`` file for a MacroModel optimization.

        The created ``.com`` file fixes all bond parameters which were
        not added during assembly. This means all bond distances, bond
        angles and torsional angles are fixed, except for cases where
        it involves a bond added during assembly of the macromolecule.

        This fixing is implemented by creating a ``.com`` file with
        various "FX" commands written within its body.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule which is to be optimized.

        Returns
        -------
        None : :class:`NoneType`

        """

        logger.debug('Creating .com file for "{}".'.format(mol.name))

        # This is the body of the ``.com`` file. The line that begins
        # and ends with exclamation lines is replaced with the various
        # commands that fix bond distances and angles.
        line1 = ('MMOD', 0, 1, 0, 0, 0, 0, 0, 0)
        line2 = ('FFLD', self.force_field, 1, 0, 0, 1, 0, 0, 0)
        line3 = ('BGIN', 0, 0, 0, 0, 0, 0, 0, 0)
        line4 = ('READ', 0, 0, 0, 0, 0, 0, 0, 0)
        line5 = ('CONV', 2, 0, 0, 0, self.minimum_gradient, 0, 0, 0)
        line6 = ('MINI', 1, 0, self.maximum_iterations, 0, 0, 0, 0, 0)
        line7 = ('END', 0, 1, 0, 0, 0, 0, 0, 0)

        com_block = "\n".join([
            self.com_line(*line1),
            self.com_line(*line2),
            self.com_line(*line3),
            self.com_line(*line4),
            '!!!BLOCK_OF_FIXED_PARAMETERS_COMES_HERE!!!',
            self.com_line(*line5),
            self.com_line(*line6),
            self.com_line(*line7)
        ])

        # Create a path for the ``.com`` file. It is the same as that
        # of the structure file but with a ``.com`` extension.
        name, ext = os.path.splitext(mol._file)

        # This function adds all the lines which fix bond distances and
        # angles into com_block.
        com_block = self.fix_params(mol, com_block)

        # Writes the .com file.
        with open(f'{name}.com', 'w') as com:
            # The first line holds the .mae file containing the
            # molecule to be optimized.
            com.write(f'{name}.mae\n')
            # The second line holds the name of the output file of the
            # optimization.
            com.write(f'{name}-out.maegz\n')
            # Next is the body of the .com file.
            com.write(com_block)

    def optimize(self, mol, conformer=-1):
        """
        Optimizes a molecule.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        conformer : :class:`int`, optional
            The conformer to use.

        Returns
        -------
        None : :class:`NoneType`

        """

        basename = str(uuid4().int)
        if self.output_dir is None:
            output_dir = basename
        else:
            output_dir = self.output_dir

        mol._file = f'{basename}.mol'

        # First write a .mol file of the molecule.
        mol.write(mol._file, conformer)
        # MacroModel requires a ``.mae`` file as input.
        self.create_mae(mol)
        # generate the ``.com`` file for the MacroModel run.
        self.generate_com(mol)
        # Run the optimization.
        self.run_bmin(mol)
        # Get the ``.maegz`` optimization output to a ``.mae``.
        self.convert_maegz_to_mae(mol)
        mol.update_from_mae(f'{basename}.mae', conformer)

        if self.restricted == 'both':
            self.restricted = False
            self.optimize(mol, conformer)
            self.restricted = 'both'

        move_generated_macromodel_files(basename, output_dir)


class MacroModelMD(_MacroModel):
    """
    Runs a MD conformer search using MacroModel.

    Attributes
    ----------
    temperature : :class:`float`
        The temperature in Kelvin at which the MD is run.

    conformers : :class:`int`
        The number of conformers sampled and optimized from the MD.

    time_step : :class:`float`
        The time step in ``fs`` for the MD.

    eq_time : :class:`float`
        The equilibriation time in ``ps`` before the MD is run.

    simulation_time : :class:`float`
        The simulation time in ``ps`` of the MD.

    """

    def __init__(self,
                 macromodel_path,
                 output_dir=None,
                 timeout=None,
                 force_field=16,
                 restricted='both',
                 temperature=300,
                 conformers=50,
                 time_step=0.1,
                 eq_time=10,
                 simulation_time=200,
                 maximum_iterations=2500,
                 minimum_gradient=0.05,
                 use_cache=False):
        """
        Runs a MD conformer search on `mol`.

        Parameters
        ----------
        macromodel_path : :class:`str`
            The full path of the Schrodinger suite within the user's
            machine. For example, on a Linux machine this may be
            something like ``'/opt/schrodinger2017-2'``.

        output_dir : :class:`str`, optional
            The name of the directory into which files generated during
            the optimization are written, if ``None`` then
            :func:`uuid.uuid4` is used.

        timeout : :class:`float`, optional
            The amount in seconds the MD is allowed to run before
            being terminated. ``None`` means there is no timeout.

        force_field : :class:`int`, optional
            The number of the force field to be used.

        restricted : :class:`bool`, optional
            If ``False`` then all bonds are optimized, not just the
            ones created during macromolecular assembly. If ``True``
            then an optimization is performed only on the bonds added
            during molecular assembly. If ``'both'`` then a restricted
            optimization is performed first, followed by a regular
            optimization.

        temperature : :class:`float`, optional
            The temperature in Kelvin at which the MD is run.

        conformers' : :class:`int`, optional
            The number of conformers sampled and optimized from the MD.

        time_step : :class:`float`, optional
            The time step in ``fs`` for the MD.

        eq_time : :class:`float`, optional
            The equilibriation time in ``ps`` before the MD is run.

        simulation_time : :class:`float`, optional
            The simulation time in ``ps`` of the MD.

        maximum_iterations : :class:`int`, optional
            The maximum number of iterations done during the
            optimization.

        minimum_gradient : :class:`float`, optional
            The gradient at which optimization is stopped.

        use_cache : :class:`bool`, optional
            If ``True`` :meth:`optimize` will not run twice on the same
            molecule and conformer.

        """

        self.temperature = temperature
        self.conformers = conformers
        self.time_step = time_step
        self.eq_time = eq_time
        self.simulation_time = simulation_time
        super().__init__(macromodel_path=macromodel_path,
                         output_dir=output_dir,
                         timeout=timeout,
                         force_field=force_field,
                         restricted=restricted,
                         maximum_iterations=maximum_iterations,
                         minimum_gradient=minimum_gradient,
                         use_cache=use_cache)

    def generate_com(self, mol):
        """
        Create a ``.com`` file for a MacroModel optimization.

        The created ``.com`` file fixes all bond parameters which were
        not added during assembly. This means all bond distances, bond
        angles and torsional angles are fixed, except for cases where
        it involves a bond added during assembly of the macromolecule.

        This fixing is implemented by creating a ``.com`` file with
        various "FX" commands written within its body.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule which is to be optimized.

        Returns
        -------
        None : :class:`NoneType`

        """

        logger.debug(f'Creating .com file for "{mol.name}".')

        # Define some short aliases to keep the following lines neat.
        temp = self.temperature
        sim_time = self.simulation_time
        tstep = self.time_step

        line1 = ('MMOD', 0, 1, 0, 0, 0, 0, 0, 0)
        line2 = ('FFLD', self.force_field, 1, 0, 0, 1, 0, 0, 0)
        line3 = ('READ', 0, 0, 0, 0, 0, 0, 0, 0)
        line4 = ('MDIT', 0, 0, 0, 0, temp, 0, 0, 0)
        line5 = ('MDYN', 0, 0, 0, 0, tstep, self.eq_time, temp, 0)
        line6 = ('MDSA', self.conformers, 0, 0, 0, 0, 0, 1, 0)
        line7 = ('MDYN', 1, 0, 0, 0, tstep, sim_time, temp, 0)
        line8 = ('WRIT', 0, 0, 0, 0, 0, 0, 0, 0)
        line9 = ('RWND', 0, 1, 0, 0, 0, 0, 0, 0)
        line10 = ('BGIN', 0, 0, 0, 0, 0, 0, 0, 0)
        line11 = ('READ', -2, 0, 0, 0, 0, 0, 0, 0)
        line12 = ('CONV', 2, 0, 0, 0, self.minimum_gradient, 0, 0, 0)
        line13 = ('MINI', 1, 0, self.maximum_iterations, 0, 0, 0, 0, 0)
        line14 = ('END', 0, 1, 0, 0, 0, 0, 0, 0)

        com_block = "\n".join([
            self.com_line(*line1),
            self.com_line(*line2),
            self.com_line(*line3),
            '!!!BLOCK_OF_FIXED_PARAMETERS_COMES_HERE!!!',
            self.com_line(*line4),
            self.com_line(*line5),
            self.com_line(*line6),
            self.com_line(*line7),
            self.com_line(*line8),
            self.com_line(*line9),
            self.com_line(*line10),
            self.com_line(*line11),
            '!!!BLOCK_OF_FIXED_PARAMETERS_COMES_HERE!!!',
            self.com_line(*line12),
            self.com_line(*line13),
            self.com_line(*line14)
        ])

        com_block = self.fix_params(mol, com_block)

        name, ext = os.path.splitext(mol._file)

        # Generate the com file containing the info for the run
        with open(f'{name}.com', 'w') as com:
            # name of the macromodel file
            com.write(f'{name}.mae\n')
            # name of the output file
            com.write(f'{name}-out.maegz\n')
            # details of the macromodel run
            com.write(com_block)

    def optimize(self, mol, conformer=-1):
        """
        Optimizes a molecule.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        conformer : :class:`int`, optional
            The conformer to use.

        Returns
        -------
        None : :class:`NoneType`

        """

        basename = str(uuid4().int)
        if self.output_dir is None:
            output_dir = basename
        else:
            output_dir = self.output_dir

        mol._file = f'{basename}.mol'

        # First write a .mol file of the molecule.
        mol.write(mol._file, conformer)
        # MacroModel requires a ``.mae`` file as input.
        self.create_mae(mol)
        # Generate the ``.com`` file for the MacroModel MD run.
        self.generate_com(mol)
        # Run the optimization.
        self.run_bmin(mol)
        # Extract the lowest energy conformer into its own .mae file.
        conformer_mae = MAEExtractor(mol._file).path
        mol.update_from_mae(conformer_mae, conformer)

        if self.restricted == 'both':
            self.restricted = False
            self.optimize(mol, conformer)
            self.restricted = 'both'

        move_generated_macromodel_files(basename, output_dir)



"""
Defines MOPAC optimizers.

"""

import os
import subprocess as sp
import psutil
import logging
from uuid import uuid4

from .optimizers import Optimizer

logger = logging.getLogger(__name__)


class MOPAC(Optimizer):
    """
    Uses MOPAC to optimize molcules.

    Attributes
    ----------
    mopac_path : :class:`str`
        The full path to the MOPAC suite on the user's
        machine. For example, in a default MacOS installation the
        folder will probably be something like
        ``/opt/mopac/MOPAC2016.exe``.

    hamiltonian : :class:`str`, optional
        A series of different methods can be selected:
        ``'PM7'``, ``'PM6'``, ``'AM1'``, ``'CIS'``
        ``'(CISD, CISDT)'``, ``'MNDO'``, ``'RM1'``, etc.

        ``'PM7'`` is the latest version of the reparametrization of
        NDDO theory, where all the atomic and diatomic
        parameters were re-optimized / updated from PM6 [#]_.

    method : :class:`str`, optional
        The default calculation consists of a geometry
        optimization. You can run single point calculations
        (SCF) or transition search algorithms (TS). Refer to
        the MOPAC website for specific keywords.

    minimum_gradient : :class:`float`, optional
        The gradient at which the geometry optimization reaches
        the convergence criteria (``kcal`` / ``mol`` /
        ``Angstrom``). For small system and high precision
        work, ``0.01`` is recommended.

    eps : :class:`float`, optional
        Sets the dielectric constant for the solvent. Presence
        of this keyword will cause the COSMO (Conductor-like
        Screening Model) method to be used to approximate the
        effect of a solvent model surrounding the molecule.
        Solvents with a low dielectric constant are not likely
        to work well with this model. ``0`` means that the
        dielectric constant is not included in the calculation.
        ``80.1`` can be used to model a water environment at
        room temperature.

    charge : :class:`float`, optional
        The charge of the system.

    fileout : :class:`str`, optional
        Determines the output file type.

    timeout : :class:`float`, optional
        The amount in seconds the optimization is allowed to
        run before being terminated. The default value is ``2``
        days or ``172,800`` seconds. ``None`` means there
        is no timeout.

    References
    ----------
    .. [#] http://openmopac.net/PM7_accuracy/PM7_accuracy.html

    """

    def __init__(self,
                 mopac_path,
                 hamiltonian='PM7',
                 method='OPT',
                 minimum_gradient=0.01,
                 eps=80.1,
                 charge=0,
                 fileout='PDBOUT',
                 timeout=172800,
                 use_cache=False):
        """
        Initializes a :class:`MOPAC` instance.

        Parameters
        ----------
        mopac_path : :class:`str`
            The full path to the MOPAC suite on the user's
            machine. For example, in a default MacOS installation the
            folder will probably be something like
            ``/opt/mopac/MOPAC2016.exe``.

        hamiltonian : :class:`str`, optional
            A series of different methods can be selected:
            ``'PM7'``, ``'PM6'``, ``'AM1'``, ``'CIS'``
            ``'(CISD, CISDT)'``, ``'MNDO'``, ``'RM1'``, etc.

            ``'PM7'`` is the latest version of the reparametrization of
            NDDO theory, where all the atomic and diatomic
            parameters were re-optimized / updated from PM6 [#]_.

        method : :class:`str`, optional
            The default calculation consists of a geometry
            optimization. You can run single point calculations
            (SCF) or transition search algorithms (TS). Refer to
            the MOPAC website for specific keywords.

        minimum_gradient : :class:`float`, optional
            The gradient at which the geometry optimization reaches
            the convergence criteria (``kcal`` / ``mol`` /
            ``Angstrom``). For small system and high precision
            work, ``0.01`` is recommended.

        eps : :class:`float`, optional
            Sets the dielectric constant for the solvent. Presence
            of this keyword will cause the COSMO (Conductor-like
            Screening Model) method to be used to approximate the
            effect of a solvent model surrounding the molecule.
            Solvents with a low dielectric constant are not likely
            to work well with this model. ``0`` means that the
            dielectric constant is not included in the calculation.
            ``80.1`` can be used to model a water environment at
            room temperature.

        charge : :class:`float`, optional
            The charge of the system.

        fileout : :class:`str`, optional
            Determines the output file type.

        timeout : :class:`float`, optional
            The amount in seconds the optimization is allowed to
            run before being terminated. The default value is ``2``
            days or ``172,800`` seconds. ``None`` means there
            is no timeout.

        use_cache : :class:`bool`, optional
            If ``True`` :meth:`optimize` will not run twice on the same
            molecule and conformer.

        References
        ----------
        .. [#] http://openmopac.net/PM7_accuracy/PM7_accuracy.html

        """

        self.mopac_path
        self.hamiltonian = hamiltonian
        self.method = method
        self.minimum_gradient = minimum_gradient
        self.eps = eps
        self.charge = charge
        self.fileout = fileout
        self.timeout = timeout
        super().__init__(use_cache=use_cache)

    def optimize(self, mol, conformer=-1):
        """
        Optimizes a molecule.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        conformer : :class:`int`, optional
            The conformer to use.

        Returns
        -------
        None : :class:`NoneType`

        """

        mol._file = f'{uuid4().int}.mol'

        # First write a .mol file of the molecule.
        mol.write(mol._file, conformer=conformer)
        # MOPAC requires a ``.mop`` file as input. This creates a
        # ``.mop`` file holding the molecule.
        self.generate_mop(mol, conformer)
        # Run the optimization.
        self.run_mopac(mol)
        # Update the mol with the generated .pdb file.
        ofile = mol._file.replace('.mol', '.pdb')
        mol.update_from_pdb(ofile, conformer=conformer)

    def run_mopac(self, mol, mopac_path, settings, timeout=7200):
        name, ext = os.path.splitext(mol._file)
        mop_file = f'{name}.mop'

        logger.info(f'Running MOPAC on "{mol.name}".')

        # To run MOPAC a command is issued to the console via
        # ``subprocess.Popen``. The command is the full path of the
        # ``mopac`` program.
        file_root, ext = os.path.splitext(mop_file)
        opt_cmd = [mopac_path, file_root]
        opt_proc = psutil.Popen(opt_cmd,
                                stdout=sp.PIPE,
                                stderr=sp.STDOUT,
                                universal_newlines=True)

        try:
            proc_out, _ = opt_proc.communicate(timeout=timeout)
        except sp.TimeoutExpired:
            logger.warning(
                f'Minimization took too long and was terminated '
                f'by force - {mol.name}'
            )
            self.kill_mopac(mol)

        return

    def kill_mopac(self, mol):
        """
        Kills an in-progress MOPAC run.

        To kill a MOPAC run, a file with the molecule's name and a
        ``.end`` extension is written.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule being optimized.

        Returns
        -------
        None : :class:`NoneType`

        """

        name, ext = os.path.splitext(mol._file)
        end_file = name + '.end'

        with open(end_file, 'w') as end:
            end.write('SHUT')

    def mop_header(self):
        """
        Returns the header of the ``.mop`` file.

        Returns
        -------
        :class:`str`
            String containing all the MOPAC keywords correctly
            formatted for the input file.

        """

        if self.method == 'OPT':
            grad = f'GNORM={self.minimum_gradient}'
        else:
            grad = ''

        return (
            f'{self.hamiltonian} {self.method} {self.method} {grad} '
            f'EPS={self.eps} CHARGE={self.charge} {self.fileout} LET '
        )

    def generate_mop(self, mol, conformer):
        """
        Creates the ``.mop`` file for the optimization.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule which is to be optimized. Its molecular
            structure file is converted to a ``.mop`` file.

        conformer : :class:`int`
            The conformer of the molecule to be used.

        Returns
        -------
        :class:`str`
            The full path of the newly created ``.mop`` file.

        """

        logger.info(f'Creating ".mop" file - {mol.name}.')
        name, ext = os.path.splitext(mol._file)
        mop_file = f'{name}.mop'

        # Generate the mop file containing the MOPAC run info.
        with open(mop_file, 'w') as mop:
            # Line for the run info.
            mop.write(f'{self.mop_header()}\n')
            # Line with the name of the molecule.
            mop.write(f'{name}\n\n')

            # Write the structural info.
            atom_block = ''
            for i in range(mol.mol.GetNumAtoms()):
                elem = mol.atom_symbol(i)
                x, y, z = mol.atom_coords(i, conformer)
                atom_block += (
                    f'{elem}   {x}   +1  {y}   +1  {z}   +1 \n'
                )

            mop.write(atom_block)

        return mop_file



"""
Defines energy calculators.

Energy calculators are objects which calculate the energy of molecules.
Each :class:`EnergyCalculator` is initialized with some settings and
calculates the energy of a molecule with
:meth:`~EnergyCalculator.energy`.

.. code-block:: python

    # Energy calculators work with any Molecule objects, such as
    # StructUnit, Polymer, Cage, etc.
    mol1 = StructUnit(...)
    mol2 = Cage(...)
    mol3 = Polymer(...)

    # Create the energy calculator.
    mmff = MMFFEnergy()

    # Calculate energies of various molecules.
    mol1_energy = mmff.energy(mol1)
    # We can optionally specify a conformer.
    mol2_energy = mmff.energy(mol2, conformer=1)
    mol3_energy = mmff.energy(mol3)

By default, calling :meth:`~EnergyCalculator.energy` twice on the
will calculate the energy a second time. However, we can use the
:attr:`~EnergyCalculator.use_cache` option to prevent recalculations
when the same molecule and conformer are given to the same energy
calculator a second time.

.. code-block:: python

    caching_mmff = MMFFEnergy(use_cache=True)
    # Calculate the energy the first time.
    energy1 = caching_mmff.energy(mol1)
    # The second time, the enegy is returned directly from memory, a
    # second calculation is not run.
    energy2 = caching_mmff.enegy(mol2)

.. _`adding energy calculators`:

Extending stk: Making new energy calculators.
---------------------------------------------

New energy calculators can be made by simply making a class which
inherits the :class:`EnergyCalculator` class. In addition to this,
the new class must define a :meth:`~EnergyCalculator.energy` method.
The method must take 2 arguments, a mandatory `mol` argument and an
optional `conformer` argument. The method will then calculate and
return the energy. There are no requirements regarding how it should
go about calculating the energy.

"""

import rdkit.Chem.AllChem as rdkit
import logging
from functools import wraps


logger = logging.getLogger(__name__)


class EnergyError(Exception):
    """
    Indicates a failed energy calculation.

    """

    ...


def _add_cache_use(energy):
    """
    Makes :meth:`~EnergyCalculator.energy` use the cache.

    Decorates `energy` so that before running it checks if the
    :class:`.Molecule` and conformer have already had their energy
    calculated. If so, and :attr:`~EnergyCalculator.use_cache` is
    ``True``, then the energy value in the
    :attr:`~EnergyCalculator.cache` is returned.

    Parameters
    ----------
    energy : :class:`function`
        A function which is to have cache use added to it.

    Returns
    -------
    :class:`function`
        The decorated function.

    """

    @wraps(energy)
    def inner(self, mol, conformer=-1):
        key = (mol.key, conformer)
        if self.use_cache and key in self.cache:
            logger.info(
                f'Using cached energy value with '
                f'"{mol.name}" conformer {conformer}.'
            )
            return self.cache[key]
        else:
            e = energy(self, mol, conformer)
            if self.use_cache:
                self.cache[key] = e
            return e

    return inner


class EnergyCalculator:
    """
    Calculate the energy of molecules.

    Attributes
    ----------
    cache : :class:`dict`
        A :class:`dict` hodling calculated energy values of the form

        .. code-block:: python

            cache = {
                (mol1, conformer1): 12.2,
                (mol1, conformer2): 2.0,
                (mol2, conformer2): 321.12
            }

        which holds every :class:`Molecule` and conformer whose energy
        was calculated by the :class:`EnergyCalculator`. Here ``mol1``
        and ``mol2`` are :class:`.Molecule` objects and ``conformer1``
        and ``conformer2`` are of type :class:`int`, and are the ids
        of the conformers passed to :meth:`energy`.

    use_cache : :class:`bool`
        If ``True`` :meth:`energy` will not run twice on the same
        molecule and conformer, but will instead return the previously
        calculated value.

    """

    def __init__(self, use_cache=False):
        """
        Initializes a :class:`EnergyCalculator` instance.

        Parameters
        ----------
        use_cache : :class:`bool`, optional
            If ``True`` :meth:`energy` will not run twice on the same
            molecule and conformer, but will instead return the previously
            calculated value.

        """

        self.cache = {}
        self.use_cache = use_cache

    def __init_subclass__(cls, **kwargs):
        cls.energy = _add_cache_use(cls.energy)
        return super().__init_subclass__(**kwargs)

    def energy(self, mol, conformer=-1):
        """
        Calculates the energy of `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The :class:`.Molecule` whose energy is to be calculated.

        conformer : :class:`int`, optinal
            The conformer of `mol` to use.

        Returns
        -------
        :class:`float`
            The energy.

        """

        raise NotImplementedError()


class FormationEnergy(EnergyCalculator):
    """
    Calculates the formation energy of a molecule.

    Attributes
    ----------
    energy_calculator : :class:`EnergyCalculator`
        The :class:`EnergyCalculator` used to calculate the energy of
        all reactant and product molecules.

    reactants : :class:`list` of :class:`.Molecule`
        The reactants. If there are multiples of the same reactant
        then it must appear multiple times in this :class:`list`.

    products : :class:`list` of :class:`.Molecule`
        The molecules which are produced as a result of the formation
        reaction. This :class:`list` must omit the :class:`.Molecule`
        passed to :meth:`energy`. If there are multiples of the same
        product, it must appear multiple times in this :class:`list`.

    reactant_conformers : :class:`list` of :class:`int`
        The conformer ids of the molecules in :attr:`reactants` to use.

    product_conformers : :class:`list` of :class:`int`
        The conformer ids of the molecules in :attr:`products` to use.

    Examples
    --------
    .. code-block:: python

        # Make the molecules needed to calculate formation energy.
        water = StructUnit.smiles_init('[H]O[H]')
        bb1 = StructUnit2(...)
        bb2 = StructUnit3(...)
        cage = Cage([bb1, bb2], FourPlusSix())

        # Make the energy calculator used to calculate energies.
        uff_energy = UFFEnergy()

        # Make the formation energy calculator.
        formation = FormationEnergy(
            energy_calculator=uff_energy,
            reactants=[bb1]*6 + [bb2]*4,
            products=[water]*cage.topology.bonds_made
        )

        # Calculate the formation energy.
        formation_energy = formation.energy(cage)

    """

    def __init__(self,
                 energy_calculator,
                 reactants,
                 products,
                 reactant_conformers=None,
                 product_conformers=None,
                 use_cache=False):
        """
        Initializes a :class:`FormationEnergy` instance.

        Parameters
        ----------
        energy_calculator : :class:`EnergyCalculator`
            The :class:`EnergyCalculator` used to calculate the energy
            of all reactant and product molecules.

        reactants : :class:`list` of :class:`.Molecule`
            The reactants. If there are multiples of the same reactant
            then it must appear multiple times in this :class:`list`.

        products : :class:`list` of :class:`.Molecule`
            The molecules which are produced as a result of the
            formation reaction. This :class:`list` must omit the
            :class:`.Molecule` passed to :meth:`energy`. If there are
            multiples of the same product, it must appear multiple
            times in this :class:`list`.

        reactant_conformers : :class:`list` of :class:`int`, optional
            The conformer ids of the molecules in :attr:`reactants` to
            use.

        product_conformers : :class:`list` of :class:`int`, optional
            The conformer ids of the molecules in :attr:`products` to
            use.

        use_cache : :class:`bool`, optional
            If ``True`` :meth:`energy` will not run twice on the same
            molecule and conformer, but will instead return the
            previously calculated value.

        """

        if reactant_conformers is None:
            reactant_conformers = [-1 for i in range(len(reactants))]

        if product_conformers is None:
            product_conformers = [-1 for i in range(len(products))]

        self.energy_calculator = energy_calculator
        self.reactants = reactants
        self.products = products
        self.reactant_conformers = reactant_conformers
        self.product_conformers = product_conformers
        super().__init__(use_cache=use_cache)

    def energy(self, mol, conformer=-1):
        """
        Calculates the formation energy of `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The :class:`.Molecule` whose formation energy is to be
            calculated.

        conformer : :class:`int`, optinal
            The conformer of `mol` to use.

        Returns
        -------
        :class:`float`
            The formation energy.

        """

        products = zip(self.products,  self.product_conformers)
        product_energy = sum(
            self.energy_calculator.energy(product, conformer)
            for product, conformer in products
        )
        product_energy += self.energy_calculator.energy(mol, conformer)

        reactants = zip(self.reactants, self.reactant_conformers)
        reactant_energy = sum(
            self.energy_calculator.energy(reactant, conformer)
            for reactant, conformer in reactants
        )

        return product_energy - reactant_energy


class MMFFEnergy(EnergyCalculator):
    """
    Uses the MMFF force field to calculate energies.

    Examples
    --------
    .. code-block:: python

        # Create a molecules whose energy we want to know.
        mol1 = StructUnit.smiles_init('CCCNCCCN')
        mol2 = Polymer(...)
        mol3 = Cage(...)

        # Create the energy calculator.
        mmff = MMFFEnergy()

        # Calculate the energies.
        energy1 = mmff.energy(mol1)
        energy2 = mmff.energy(mol2)
        energy3 = mmff.energy(mol3)

    """

    def energy(self, mol, conformer=-1):
        """
        Calculates the energy of `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The :class:`.Molecule` whose energy is to be calculated.

        conformer : :class:`int`, optinal
            The conformer of `mol` to use.

        Returns
        -------
        :class:`float`
            The energy.

        """

        rdkit.GetSSSR(mol.mol)
        mol.mol.UpdatePropertyCache()
        ff = rdkit.MMFFGetMoleculeForceField(
            mol.mol,
            rdkit.MMFFGetMoleculeProperties(mol.mol),
            confId=conformer
        )

        return ff.CalcEnergy()


class UFFEnergy(EnergyCalculator):
    """
    Uses the UFF force field to calculate energies.

    Examples
    --------
    .. code-block:: python

        # Create a molecules whose energy we want to know.
        mol1 = StructUnit.smiles_init('CCCNCCCN')
        mol2 = Polymer(...)
        mol3 = Cage(...)

        # Create the energy calculator.
        uff = UFFEnergy()

        # Calculate the energies.
        energy1 = uff.energy(mol1)
        energy2 = uff.energy(mol2)
        energy3 = uff.energy(mol3)

    """

    def energy(self, mol, conformer=-1):
        """
        Calculates the energy of `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The :class:`.Molecule` whose energy is to be calculated.

        conformer : :class:`int`, optinal
            The conformer of `mol` to use.

        Returns
        -------
        :class:`float`
            The energy.

        """

        mol.mol.UpdatePropertyCache()
        ff = rdkit.UFFGetMoleculeForceField(mol.mol, confId=conformer)
        return ff.CalcEnergy()



"""
Defines optimizers.

Optimizers are objects used to optimize molecules. Each optimizer is
initialized with some settings and used to optimize a molecule
with :meth:`~Optimizer.optimize`.

.. code-block:: python

    import rdkit.Chem.AllChem as rdkit
    mol = StructUnit2.smiles_init('NCCCN', ['amine'])
    mmff = MMFF()
    mmff.optimize(mol)

    # Optionally, a conformer can be provided.
    mmff.optimize(mol, conformer=2)

    # Optimizers also work with MacroMolecule objects.
    polymer = Polymer([mol], Linear('A', [0], n=3))
    etkdg = RDKitEmbedder(rdkit.ETKDG())
    etkdg.optimize(polymer)

Sometimes it is desirable to chain multiple optimizations, one after
another. For example, before running an optimization, it may be
desirable to embed a molecule first, to generate an initial structure.
:class:`OptimizerSequence` may be used for this.

.. code-block:: python

    # Create a new optimizer which chains the previously defined
    # mmff and etkdg optimizers.
    optimizer_sequence = OptimizerSequence(etkdg, mmff)

    # Run each optimizer in sequence.
    optimizer_sequence.optimize(polymer)

By default, running :meth:`Optimizer.optimize` twice in a row will
perform an optimization a second time on a molecule. If we want to
skip optimizations on molecules which have already been optimized
we can use the :attr:`use_cache` flag.

.. code-block:: python

    caching_etkdg = RDKitEmbedder(rdkit.ETKDG(), use_cache=True)
    # First optimize call runs an optimization.
    caching_etkdg.optimize(polymer)
    # Second call does nothing.
    caching_etkdg.optimize(polymer)

Caching is done on a per :class:`Optimizer` basis. Just because the
molecule has been cached by one :class:`Optimizer` instance does not
mean that a different :class:`Optimizer` instance will no longer
optimize the molecule.

.. _`adding optimizers`:

Extending stk: Making new optimizers.
-------------------------------------

New optimizers can be made by simply making a class which inherits the
:class:`Optimizer` class. In addition to this, the new class must
define a :meth:`~Optimize.optimize` method. The method must take 2
arguments a mandatory `mol` argument and an optional `conformer`
argument. :meth:`~Optimizer.optimize` will take a molecule and change
its structure however it likes. Beyond this there are no requirements.

"""

import logging
import numpy as np
import rdkit.Chem.AllChem as rdkit
import warnings
from functools import wraps

logger = logging.getLogger(__name__)


def _add_cache_use(optimize):
    """
    Makes :meth:`~Optimizer.optimize` use the :attr:`~Optimizer.cache`.

    Decorates `optimize` so that before running it checks if the
    :class:`.Molecule` and conformer have already been optimized by the
    optimizer. If so, and :attr:`~Optimizer.use_cache` is ``True``,
    then the molecule is skipped and no optimization is performed.

    Parameters
    ----------
    optimize : :class:`function`
        A function which is to have skipping added to it.

    Returns
    -------
    :class:`function`
        The decorated function.

    """

    @wraps(optimize)
    def inner(self, mol, conformer=-1):
        key = (mol.key, conformer)
        if self.use_cache and key in self.cache:
            logger.info(
                f'Skipping optimization on '
                f'"{mol.name}" conformer {conformer}.'
            )
        else:
            optimize(self, mol, conformer)
            if self.use_cache:
                self.cache.add(key)

    return inner


class Optimizer:
    """
    A base class for optimizers.

    Attributes
    ----------
    cache : :class:`set`
        A :class:`set` of the form

        .. code-block:: python

            cache = {
                (mol1, conformer1),
                (mol1, conformer2),
                (mol2, conformer2)
            }

        which holds every :class:`Molecule` and conformer optimized
        by the :class:`Optimizer`. Here ``mol1`` and ``mol2`` are
        :class:`.Molecule` objects and ``conformer1`` and
        ``conformer2`` are :class:`int`, which are the ids of the
        optimized conformers of the molecules.

    use_cache : :class:`bool`
        If ``True`` :meth:`optimize` will not run twice on the same
        molecule and conformer.

    """

    def __init__(self, use_cache=False):
        """
        Initializes an :class:`Optimizer`.

        Parameters
        ----------
        use_cache : :class:`bool`, optional
            If ``True`` :meth:`optimize` will not run twice on the same
            molecule and conformer.

        """

        self.cache = set()
        self.use_cache = use_cache

    def __init_subclass__(cls, **kwargs):
        cls.optimize = _add_cache_use(cls.optimize)
        return super().__init_subclass__(**kwargs)

    def optimize(self, mol, conformer=-1):
        """
        Optimizes a molecule.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        conformer : :class:`int`, optional
            The conformer to use.

        Returns
        -------
        None : :class:`NoneType`

        """

        raise NotImplementedError()


class OptimizerSequence(Optimizer):
    """
    Applies optimizers in sequence.

    Attributes
    ----------
    optimizers : :class:`tuple` of :class:`Optimizer`
        A number of optimizers, each of which gets applied to a
        molecule, based on the order in this :class:`tuple`.

    Examples
    --------
    Let's say we want to embed a molecule with ETKDG first and then
    minimize it with the MMFF force field.

    .. code-block:: python

        import rdkit.Chem.AllChem as rdkit
        mol = StructUnit.smiles_init('NCCNCCN', ['amine'])
        etkdg = RDKitEmbedder(rdkit.ETKDG())
        mmff = RDKitForceField(rdkit.MMFFOptimizeMolecule)
        optimizer = OptimizerPipeline(etkdg, mmff)
        optimizer.optimize(mol)

    """

    def __init__(self, *optimizers, use_cache=False):
        """
        Initializes a :class:`OptimizerSequence` instance.

        Parameters
        ----------
        *optimizers : :class:`Optimizer`
            A number of optimizers, each of which gets applied to a
            molecule, based on the order given.

        use_cache : :class:`bool`, optional
            If ``True`` :meth:`optimize` will not run twice on the same
            molecule and conformer.

        """

        self.optimizers = optimizers
        super().__init__(use_cache=use_cache)

    def optimize(self, mol, conformer=-1):
        """
        Chains multiple :class:`Optimizer` instances together.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        conformer : :class:`int`, optional
            The conformer to use.

        Returns
        -------
        None : :class:`NoneType`

        """

        for optimizer in self.optimizers:
            cls_name = optimizer.__class__.__name__
            logger.info(f'Using {cls_name} on "{mol.name}".')
            optimizer.optimize(mol, conformer)


class CageOptimizerSequence(Optimizer):
    """
    Applies :class:`Optimizer` objects to a :class:`.Cage`.

    Before each :class:`Optimizer` in the sequence is applied to the
    :class:`.Cage`, it is checked to see if it is collapsed. If it is
    collapsed, the optiimzation sequence ends immediately.

    Attributes
    ----------
    optimizers : :class:`tuple` of :class:`Optimizer`
        The :class:`Optimizer` objects which are used to optimize a
        :class:`.Cage` molecule.

    """

    def __init__(self, *optimizers):
        """
        Initializes a :class:`CageOptimizerSequence` instance.

        Parameters
        ----------
        *optimizers : :class:`Optimizer`
            The :class:`Optimizers` used in sequence to optimize
            :class:`.Cage` molecules.

        """

        self.optimizers = optimizers
        # "optimizers" should toggle use_cache for themselves.
        super().__init__(use_cache=False)

    def optimize(self, mol, conformer=-1):
        """
        Optimizes a molecule.

        Parameters
        ----------
        mol : :class:`.Cage`
            The cage to be optimized.

        conformer : :class:`int`, optional
            The conformer to use.

        Returns
        -------
        None : :class:`NoneType`

        """

        for optimizer in self.optimizers:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                windows = mol.windows(conformer)

            logger.debug(f'Windows found: {windows}.')
            expected_windows = mol.topology.n_windows
            if windows is None or len(windows) != expected_windows:
                logger.info(
                    f'"{mol.name}" is collapsed, exiting early.'
                )
                return

            cls_name = optimizer.__class__.__name__
            logger.info(f'Using {cls_name} on "{mol.name}".')
            optimizer.optimize(mol, conformer)


class NullOptimizer(Optimizer):
    """
    Does not perform optimizations.

    """

    def optimize(self, mol, conformer=-1):
        """
        Does not optimize a molecule.

        This function just returns immediately without changing the
        molecule.

        Parameters
        ----------
        mol : :class:`.Molecule`
            A molecule.

        conformer : :class:`int`, optional
            A conformer.

        Returns
        -------
        None : :class:`NoneType`

        """

        return


class TryCatchOptimizer(Optimizer):
    """
    Try to optimize with a :class:`Optimizer`, use another on failure.

    Attributes
    ----------
    try_optimizer : :class:`Optimizer`
        The optimizer which is used initially to try and optimize a
        :class:`.Molecule`.

    catch_optimizer : :class:`Optimizer`
        If :attr:`try_optimizer` raises an error, this optimizer is
        run on the :class:`.Molecule` instead.

    Examples
    --------
    .. code-block:: python

        # Create some molecules to optimize.
        mol1 = StructUnit2.smiles_init('NCCN', []'amine'])
        mol2 = StructUnit.smiles_init('CCCCC')
        mol3 = StructUnit.smiles_init('O=CCCN')

        # Create an optimizer which may fail.
        uff = UFF()

        # Create a backup optimizer.
        mmff = MMFF()

        # Make an optimizer which tries to run raiser and if that
        # raises an error, will run mmff on the molecule instead.
        try_catch = TryCatchOptimizer(try_optimizer=uff,
                                      catch_optimizer=mmff)

        # Optimize the molecules. In each case if the optimization with
        # UFF fails, MMFF is used to optimize the molecule instead.
        try_catch.optimize(mol1)
        try_catch.optimize(mol2)
        try_catch.optimzie(mol3)


    """

    def __init__(self, try_optimizer, catch_optimizer):
        self.try_optimizer = try_optimizer
        self.catch_optimizer = catch_optimizer
        # try and catch optimizers should toggle use_cache themselves.
        super().__init__(use_cache=False)

    def optimize(self, mol, conformer=-1):
        """
        Optimizes a molecule.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        conformer : :class:`int`, optional
            The conformer to use.

        Returns
        -------
        None : :class:`NoneType`

        """

        try:
            return self.try_optimizer.optimize(mol, conformer)
        except Exception:
            try_name = self.try_optimizer.__class__.__name__
            catch_name = self.catch_optimizer.__class__.__name__
            logger.error(
                f'{try_name} failed, trying {catch_name}.',
                exc_info=True
            )
            return self.catch_optimizer.optimize(mol, conformer)


class RaisingOptimizerError(Exception):
    ...


class RaisingOptimizer(Optimizer):
    """
    Raises and optimizes at random.

    This optimizer is used for debugging to simulate optimization
    functions which sometimes completes successfully and sometimes
    randomly fails.

    Attributes
    ----------
    optimizer : :class:`Optimizer`
        When the optimizer does not fail, it uses this
        :class:`Optimizer` to optimize molecules.

    fail_chance : :class:`float`
        The probability that the optimizer will raise an error each
        time :meth:`optimize` is used.

    Examples
    --------
    .. code-block:: python

        mol = StructUnit.smiles_init('NCCNCCN', ['amine'])
        etkdg = RDKitEmbedder(rdkit.ETKDG())
        partial_raiser = RaisingOptimizer(etkdg, fail_chance=0.75)
        # 75 % chance an error will be raised by calling optimize.
        partial_raiser.optimize(mol)

    """

    def __init__(self, optimizer, fail_chance=0.5):
        """
        Initializes :class:`PartialRaiser`.

        Parameters
        ----------
        optimizer : :class:`Optimizer`
            When the optimizer does not fail, it uses this
            :class:`Optimizer` to optimize molecules.

        fail_chance : :class:`float`, optional
            The probability that the optimizer will raise an error each
            time :meth:`optimize` is used.

        """

        self.optimizer = optimizer
        self.fail_chance = fail_chance
        # "optmizer" should toggle use_cache for itself.
        super().__init__(use_cache=False)

    def optimize(self, mol, conformer=-1):
        """
        Optimizes a molecule.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        conformer : :class:`int`, optional
            The conformer to use.

        Returns
        -------
        None : :class:`NoneType`

        Raises
        ------
        :class:`RaisingOptimizerError`
            This error is raised at random.

        """

        if np.random.rand() < self.fail_chance:
            raise RaisingOptimizerError('Used RaisingOptimizer.')
        return self.optimizer.optimize(mol)


class MMFF(Optimizer):
    """
    Use the MMFF force field to optimize molecules.

    Examples
    --------
    .. code-block:: python

        mol = StructUnit.smiles_init('NCCNCCN', ['amine'])
        mmff = MMFF()
        mmff.optimize(mol)

    """

    def optimize(self, mol, conformer=-1):
        """
        Optimizes a molecule with the MMFF force field.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        conformer : :class:`int`, optional
            The conformer to use.

        Returns
        -------
        None : :class:`NoneType`

        """

        if conformer == -1:
            conformer = mol.mol.GetConformer(conformer).GetId()

        # Needs to be sanitized to get force field params.
        rdkit.SanitizeMol(mol.mol)
        rdkit.MMFFOptimizeMolecule(mol.mol, confId=conformer)


class UFF(Optimizer):
    """
    Use the UFF force field to optimize molecules.

    Examples
    --------
    .. code-block:: python

        mol = StructUnit.smiles_init('NCCNCCN', ['amine'])
        uff = UFF()
        uff.optimize(mol)

    """

    def optimize(self, mol, conformer=-1):
        """
        Optimizes a molecule with the UFF force field.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        conformer : :class:`int`, optional
            The conformer to use.

        Returns
        -------
        None : :class:`NoneType`

        """

        if conformer == -1:
            conformer = mol.mol.GetConformer(conformer).GetId()

        # Needs to be sanitized to get force field params.
        rdkit.SanitizeMol(mol.mol)
        rdkit.UFFOptimizeMolecule(mol.mol, confId=conformer)


class RDKitEmbedder(Optimizer):
    """
    Uses :func:`rdkit.EmbedMolecule` to find an optimized structure.

    Attributes
    ----------
    params : :class:`rdkit.EmbedParameters`
        The parameters used for the embedding.

    Examples
    --------
    Use ETKDG [#]_ to generate an optimized structure.

    .. code-block:: python

        import rdkit.Chem.AllChem as rdkit
        mol = StructUnit.smiles_init('NCCNCCN', ['amine'])
        embedder = RDKitEmbedder(rdkit.ETKDG())
        embedder.optimize(mol)

    References
    ----------
    .. [#] http://pubs.acs.org/doi/pdf/10.1021/acs.jcim.5b00654

    """

    def __init__(self, params, use_cache=False):
        """
        Initializes a :class:`RDKitEmbedder` instance.

        Parameters
        ----------
        params : :class:`rdkit.EmbedParameters`
            The parameters used for the embedding.

        use_cache : :class:`bool`, optional
            If ``True`` :meth:`optimize` will not run twice on the same
            molecule and conformer.

        """

        self.params = params
        super().__init__(use_cache=use_cache)

    def optimize(self, mol, conformer=-1):
        """
        Optimizes a molecule.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        conformer : :class:`int`, optional
            The conformer to use.

        Returns
        -------
        None : :class:`NoneType`

        """

        if conformer == -1:
            conformer = mol.mol.GetConformer(conformer).GetId()

        conf_id = rdkit.EmbedMolecule(mol.mol, self.params)
        new_conf = rdkit.Conformer(mol.mol.GetConformer(conf_id))
        mol.mol.RemoveConformer(conf_id)
        mol.mol.RemoveConformer(conformer)
        new_conf.SetId(conformer)
        mol.mol.AddConformer(new_conf)



class ElectronicPropertyCalculator:
    def dipole_moment(self, mol, conformer=-1):
        raise NotImplementedError()

    def electron_affinity(self, mol, conformer=-1):
        raise NotImplementedError()

    def ionization_potential(self, mol, conformer=-1):
        raise NotImplementedError()



"""
Defines cage topologies from di and 4 functionalised building blocks.

"""


import numpy as np

from .base import CageTopology,  Vertex, Edge


class TwoPlusFour(CageTopology):
    """
    Sandwich cage topology from 2 and 4 functionalized building blocks.

    """

    positions_A = [Vertex(0, 0, -1), Vertex(0, 0, 1)]
    alpha, beta = positions_A
    beta.edge_plane_normal = (
        lambda scale, alpha=alpha:
            np.multiply(alpha.edge_plane_normal(scale), -1)
    )

    positions_B = [Edge(alpha, beta),
                   Edge(alpha, beta),
                   Edge(alpha, beta),
                   Edge(alpha, beta)]

    a, b, c, d = positions_B

    a.coord = np.array([2, 0, 0])
    b.coord = np.array([-2, 0, 0])
    c.coord = np.array([0, 2, 0])
    d.coord = np.array([0, -2, 0])

    n_windows = 4
    n_window_types = 1


class ThreePlusSix(CageTopology):
    """
    A cage topology from 2 and 4 functionalized building blocks.

    """
    x = 1
    positions_A = [Vertex(-2*x, -x*np.sqrt(3), 0),
                   Vertex(2*x, -x*np.sqrt(3), 0),
                   Vertex(0, x*np.sqrt(3), 0)]

    a, b, c = positions_A

    positions_B = [Edge(a, b, custom_position=True),
                   Edge(a, b, custom_position=True),
                   Edge(b, c, custom_position=True),
                   Edge(b, c, custom_position=True),
                   Edge(a, c, custom_position=True),
                   Edge(a, c, custom_position=True)]

    e1, e2, e3, e4, e5, e6 = positions_B
    for e in [e1, e3, e5]:
        e.coord = np.add(e.coord, [0, 0, x])

    for e in [e2, e4, e6]:
        e.coord = np.add(e.coord, [0, 0, -x])

    n_windows = 5
    n_window_types = 2


class FourPlusEight(CageTopology):
    """
    A cage topology from 2 and 4 functionalized building blocks.

    """

    positions_A = [Vertex(-1, -1, 0),
                   Vertex(-1, 1, 0),
                   Vertex(1, -1, 0),
                   Vertex(1, 1, 0)]

    a, b, c, d = positions_A

    positions_B = [Edge(a, b, custom_position=True),
                   Edge(a, b, custom_position=True),
                   Edge(b, d, custom_position=True),
                   Edge(b, d, custom_position=True),
                   Edge(a, c, custom_position=True),
                   Edge(a, c, custom_position=True),
                   Edge(c, d, custom_position=True),
                   Edge(c, d, custom_position=True)]

    e1, e2, e3, e4, e5, e6, e7, e8 = positions_B
    for e in [e1, e3, e5, e7]:
        e.coord = np.add(e.coord, [0, 0, 1])

    for e in [e2, e4, e6, e8]:
        e.coord = np.add(e.coord, [0, 0, -1])

    n_windows = 6
    n_window_types = 2


class FivePlusTen(CageTopology):

    c1 = np.cos(2*np.pi/5)
    c2 = np.cos(np.pi/5)
    s1 = np.sin(2*np.pi/5)
    s2 = np.sin(4*np.pi/5)

    positions_A = [Vertex(0, 1, 0),
                   Vertex(s1, c1, 0),
                   Vertex(s2, -c2, 0),
                   Vertex(-s2, -c2, 0),
                   Vertex(-s1, c1, 0)]

    a, b, c, d, e = positions_A

    positions_B = [Edge(a, b, custom_position=True),
                   Edge(a, b, custom_position=True),
                   Edge(b, c, custom_position=True),
                   Edge(b, c, custom_position=True),
                   Edge(c, d, custom_position=True),
                   Edge(c, d, custom_position=True),
                   Edge(d, e, custom_position=True),
                   Edge(d, e, custom_position=True),
                   Edge(e, a, custom_position=True),
                   Edge(e, a, custom_position=True)]

    for e in positions_B[::2]:
        e.coord = np.add(e.coord, [0, 0, 0.5])

    for e in positions_B[1::2]:
        e.coord = np.add(e.coord, [0, 0, -0.5])

    n_windows = 7
    n_window_types = 2


class SixPlusTwelve(CageTopology):
    """
    A cage topology from 2 and 4 functionalized building blocks.

    """

    positions_A = [Vertex(-1, -1, 0),
                   Vertex(-1, 1, 0),
                   Vertex(1, -1, 0),
                   Vertex(1, 1, 0),
                   Vertex(0, 0, 1),
                   Vertex(0, 0, -1)]

    a, b, c, d, e, f = positions_A

    positions_B = [Edge(a, b),
                   Edge(b, d),
                   Edge(d, c),
                   Edge(a, c),
                   Edge(e, a),
                   Edge(e, b),
                   Edge(e, c),
                   Edge(e, d),
                   Edge(f, a),
                   Edge(f, b),
                   Edge(f, c),
                   Edge(f, d)]

    n_windows = 8
    n_window_types = 1


class EightPlusSixteen(CageTopology):

    x = 2
    positions_A = a, b, c, d, e, f, g, h = [
                    Vertex(-0.5*x, 0.5*x, -0.35*x),
                    Vertex(-0.5*x, -0.5*x, -0.35*x),
                    Vertex(0.5*x, -0.5*x, -0.35*x),
                    Vertex(0.5*x, 0.5*x, -0.35*x),

                    Vertex(-x*np.sqrt(2)/2, 0, x*0.35),
                    Vertex(0, -x*np.sqrt(2)/2, x*0.35),
                    Vertex(x*np.sqrt(2)/2, 0, x*0.35),
                    Vertex(0, x*np.sqrt(2)/2, x*0.35)]

    positions_B = [Edge(b, f),
                   Edge(c, f),
                   Edge(a, e),
                   Edge(b, e),

                   Edge(c, g),
                   Edge(d, g),
                   Edge(a, h),
                   Edge(d, h),

                   Edge(a, b),
                   Edge(b, c),
                   Edge(c, d),
                   Edge(d, a),

                   Edge(e, f),
                   Edge(f, g),
                   Edge(g, h),
                   Edge(h, e)]
    n_windows = 10
    n_window_types = 2


class TenPlusTwenty(CageTopology):
    """
    A cage topology from 2 and 4 functionalized building blocks.

    """

    x = 2
    positions_A = [Vertex(-x, x, -x),
                   Vertex(-x, -x, -x),
                   Vertex(x, x, -x),
                   Vertex(x, -x, -x),

                   Vertex(-x, x, x),
                   Vertex(-x, -x, x),
                   Vertex(x, x, x),
                   Vertex(x, -x, x),

                   Vertex(0, 0, x*1.5),
                   Vertex(0, 0, -x*1.5)]

    a, b, c, d, e, f, g, h, i, j = positions_A

    positions_B = [Edge(a, c),
                   Edge(a, b),
                   Edge(b, d),
                   Edge(c, d),

                   Edge(e, g),
                   Edge(e, f),
                   Edge(f, h),
                   Edge(g, h),


                   Edge(a, e),
                   Edge(b, f),
                   Edge(c, g),
                   Edge(d, h),

                   Edge(i, e),
                   Edge(i, f),
                   Edge(i, g),
                   Edge(i, h),

                   Edge(j, a),
                   Edge(j, b),
                   Edge(j, c),
                   Edge(j, d)]

    n_windows = 12
    n_window_types = 2



from .electronic_property_calculators import (
    ElectronicPropertyCalculator
)


class MOPACElectronicProperties(ElectronicPropertyCalculator):
    def __init__(self, mopac_path):
        self.mopac_path = mopac_path

    def dipole_moment(self, mol, conformer=-1):
        """
        Calculates the dipole moment using MOPAC.

        Note that this requires MOPAC to be installed and have a
        valid license.

        Parameters
        ----------
        settings : :class:`dict`, optional
            A dictionary which maps the names of the optimization
            parameters to their values. Valid values are:

                'hamiltonian' : :class:`str` (default = ``'PM7'``
                    A series of different methods can be selected:
                    PM7, PM6, AM1, CIS (CISD, CISDT), MNDO, RM1, etc..

                    PM7 is the latest version of the reparametrization
                    of NDDO theory, where all the atomic and diatomic
                    parameters were re-optimized / updated from PM6
                    [#]_.

                'eps' : :class:`float` (default = ``80.1``)
                    Sets the dielectric constant for the solvent.
                    Presence of this keyword will cause the COSMO
                    (Conductor-like Screening Model) method to be used
                    to approximate the effect of a solvent model
                    surrounding the molecule. Solvents with a low
                    dielectric constant are not likely to work well
                    with this model. ``0`` means that the dielectric
                    constant is not included in the calculation.
                    ``80.1`` can be used to model a water environment
                    at room temperature.

                'charge' : :class:`float` (default = ``0``)
                    The charge of the system.

                'timeout' : :class:`float` (default = ``172800``)
                    The amount in seconds the calculation is allowed to
                    run before being terminated. The default value is
                    ``2`` days or ``172,800`` seconds.

        mopac_path : :class:`str`
            The full path to the MOPAC installation.

        Returns
        -------
        :class:`float`
            The calculated dipole.

        References
        ----------
        .. [#] http://openmopac.net/PM7_accuracy/PM7_accuracy.html

        """

        if settings is None:
            settings = {}

        # Define default vals for the MOPAC input
        vals = {
                'hamiltonian': 'PM7',
                'method': 'NOOPT',
                'eps': 80.1,
                'charge': 0,
                'timeout': 172800,
                }

        vals.update(settings)

        # To prevent conflicts when running this function in parallel,
        # a temporary copy of the molecular structure file is made and
        # used for mopac calculations.

        # Unique file name is generated by inserting a random int into
        # the file path.
        tmp_file = "{}.mol".format(uuid4().int)
        self.molecule.write(tmp_file)

        file_root, ext = os.path.splitext(tmp_file)

        # Generate the input file
        _create_mop(file_root, self.molecule, vals)
        # Run MOPAC
        _run_mopac(file_root, mopac_path)
        return _extract_MOPAC_dipole(file_root)

    def electron_affinity(self, mol, conformer=-1):
        """
        Calculates the electron affinity using MOPAC.

        Note that this requires MOPAC to be installed and have a
        valid license.

        Parameters
        ----------
        settings : :class:`dict`, optional
            A dictionary which maps the names of the optimization
            parameters to their values. Valid values are:

                'hamiltonian' : :class:`str` (default = ``'PM7'``
                    A series of different methods can be selected:
                    PM7, PM6, AM1, CIS (CISD, CISDT), MNDO, RM1, etc..

                    PM7 is the latest version of the reparametrization
                    of NDDO theory, where all the atomic and diatomic
                    parameters were re-optimized / updated from PM6
                    [#]_.

                'eps' : :class:`float` (default = ``80.1``)
                    Sets the dielectric constant for the solvent.
                    Presence of this keyword will cause the COSMO
                    (Conductor-like Screening Model) method to be used
                    to approximate the effect of a solvent model
                    surrounding the molecule. Solvents with a low
                    dielectric constant are not likely to work well
                    with this model. ``0`` means that the dielectric
                    constant is not included in the calculation.
                    ``80.1`` can be used to model a water environment
                    at room temperature.

                'charge' : :class:`float` (default = ``0``)
                    The charge of the system.

                'timeout' : :class:`float` (default = ``172800``)
                    The amount in seconds the calculation is allowed to
                    run before being terminated. The default value is
                    ``2`` days or ``172,800`` seconds.

        mopac_path : :class:`str`
            The full path to the MOPAC installation.

        Returns
        -------
        :class:`float`
            The calculated energy.

        References
        ----------
        .. [#] http://openmopac.net/PM7_accuracy/PM7_accuracy.html

        """

        if settings is None:
            settings = {}

        # Define default vals for the MOPAC input
        vals = {
                'hamiltonian': 'PM7',
                'method': 'NOOPT',
                'eps': 80.1,
                'charge': 0,
                'timeout': 172800,
                }
        vals.update(settings)

        # First check the energy of the neutral system
        # To prevent conflicts when running this function in parallel,
        # a temporary copy of the molecular structure file is made and
        # used for mopac calculations.

        # Unique file name is generated by inserting a random int into
        # the file path.
        tmp_file = "{}.mol".format(uuid4().int)
        self.molecule.write(tmp_file)

        file_root, ext = os.path.splitext(tmp_file)

        # Generate the input file
        _create_mop(file_root, self.molecule, vals)
        # Run MOPAC
        _run_mopac(file_root, mopac_path)

        # Extract the neutral energy
        en1 = _extract_MOPAC_en(file_root)

        # Update the settings for the anion optimization
        settings2 = {
                    'method': 'OPT',
                    'gradient': 0.01,
                    'charge': -1,
                    'fileout': 'PDBOUT'
                    }

        vals.update(settings2)

        # Now generate a new molecule
        mol2 = copy.deepcopy(self.molecule)
        # Run the mopac optimization
        mopac_opt(mol2, mopac_path, vals)
        # Extract the energy by using the self.mopac method
        vals['method'] = 'NOOPT'
        del vals['gradient']
        del vals['fileout']
        en2 = mol2.energy.mopac(mopac_path, vals)
        # Calculate the EA (eV)
        return en2 - en1

    def ionization_potential(self, mol, conformer=-1):
        """
        Calculates the ionization potential using MOPAC.

        Note that this requires MOPAC to be installed and have a
        valid license.

        Parameters
        ----------
        settings : :class:`dict`, optional
            A dictionary which maps the names of the optimization
            parameters to their values. Valid values are:

                'hamiltonian' : :class:`str` (default = ``'PM7'``
                    A series of different methods can be selected:
                    PM7, PM6, AM1, CIS (CISD, CISDT), MNDO, RM1, etc..

                    PM7 is the latest version of the reparametrization
                    of NDDO theory, where all the atomic and diatomic
                    parameters were re-optimized / updated from PM6
                    [#]_.

                'eps' : :class:`float` (default = ``80.1``)
                    Sets the dielectric constant for the solvent.
                    Presence of this keyword will cause the COSMO
                    (Conductor-like Screening Model) method to be used
                    to approximate the effect of a solvent model
                    surrounding the molecule. Solvents with a low
                    dielectric constant are not likely to work well
                    with this model. ``0`` means that the dielectric
                    constant is not included in the calculation.
                    ``80.1`` can be used to model a water environment
                    at room temperature.

                'charge' : :class:`float` (default = ``0``)
                    The charge of the system.

                'timeout' : :class:`float` (default = ``172800``)
                    The amount in seconds the calculation is allowed to
                    run before being terminated. The default value is
                    ``2`` days or ``172,800`` seconds.

        mopac_path : :class:`str`
            The full path to the MOPAC installation.

        Returns
        -------
        :class:`float`
            The calculated energy.

        References
        ----------
        .. [#] http://openmopac.net/PM7_accuracy/PM7_accuracy.html

        """

        if settings is None:
            settings = {}

        # Define default vals for the MOPAC input
        vals = {
                'hamiltonian': 'PM7',
                'method': 'NOOPT',
                'eps': 80.1,
                'charge': 0,
                'timeout': 172800,
                }
        vals.update(settings)

        # First check the energy of the neutral system
        # To prevent conflicts when running this function in parallel,
        # a temporary copy of the molecular structure file is made and
        # used for mopac calculations.

        # Unique file name is generated by inserting a random int into
        # the file path.
        tmp_file = "{}.mol".format(uuid4().int)
        self.molecule.write(tmp_file)

        file_root, ext = os.path.splitext(tmp_file)

        # Generate the input file
        _create_mop(file_root, self.molecule, vals)
        # Run MOPAC
        _run_mopac(file_root, mopac_path)

        # Extract the neutral energy
        en1 = _extract_MOPAC_en(file_root)

        # Update the settings for the cation optimization
        settings2 = {
                    'method': 'OPT',
                    'gradient': 0.01,
                    'charge': 1,
                    'fileout': 'PDBOUT'
                    }

        vals.update(settings2)

        # Now generate a new molecule
        mol2 = copy.deepcopy(self.molecule)
        # Run the mopac optimisation
        mopac_opt(mol2, mopac_path, vals)
        # Extract the energy by using the self.mopac method
        vals['method'] = 'NOOPT'
        del vals['gradient']
        del vals['fileout']
        en2 = mol2.energy.mopac(mopac_path, vals)
        # Calculate the IP (eV)
        return en2 - en1


def _run_mopac(file_root, mopac_path, timeout=3600):

    mop_file = file_root + '.mop'

    logger.info(f'Running MOPAC - {file_root}.')

    # To run MOPAC a command is issued to the console via
    # ``subprocess.Popen``. The command is the full path of the
    # ``mopac`` program.
    file_root, ext = os.path.splitext(mop_file)
    opt_cmd = [mopac_path, file_root]
    opt_proc = psutil.Popen(opt_cmd, stdout=sp.PIPE,
                            stderr=sp.STDOUT,
                            universal_newlines=True)

    try:
        if timeout:
            proc_out, _ = opt_proc.communicate(timeout=timeout)
        else:
            proc_out, _ = opt_proc.communicate()
    except sp.TimeoutExpired:
        logger.info(('Minimization took too long and was terminated '
                     'by force - {}').format(file_root))
        _kill_mopac(file_root)


def _kill_mopac(file_root):
    """
    Stops an in-progress MOPAC run.

    To kill a MOPAC run, a file with the molecule's name and a ``.end``
    extension is written.

    Parameters
    ----------
    file_root : :class:`str`
        The molecule's name.

    Returns
    -------
    None : :class:`NoneType`

    """
    end_file = file_root + '.end'

    with open(end_file, 'w') as end:
        end.write('SHUT')


def _mop_line(settings):
    """
    Formats `settings` into a MOPAC input string.

    Parameters
    ----------
    settings : :class:`dict`
        Dictionary defined in :func:`mopac_opt`, where all the run
        details are defined.

    Returns
    -------
    :class:`str`
        String containing all the MOPAC keywords correctly formatted
        for the input file.

    """

    # Generate an empty string
    mopac_run_str = ""

    # Add Hamiltonian info
    mopac_run_str = mopac_run_str + settings['hamiltonian']
    # Forcing a single point calculation
    mopac_run_str = mopac_run_str + ' NOOPT '
    # Add EPS info
    eps_info = ' EPS={} '.format(settings['eps'])
    mopac_run_str = mopac_run_str + eps_info
    # Add Charge info
    charge_info = ' CHARGE={} '.format(settings['charge'])
    mopac_run_str = mopac_run_str + charge_info

    return mopac_run_str


def _create_mop(file_root, molecule, settings):
    """
    Creates the ``.mop`` file holding the molecule to be optimized.

    Parameters
    ----------
    mol : :class:`.Molecule`
        The molecule which is to be optimized. Its molecular
        structure file is converted to a ``.mop`` file. The original
        file is also kept.

    settings : :class:`dict`
        Dictionary defined by the MOPAC methods, where all the run
        details are defined.

    Returns
    -------
    :class:`str`
        The full path of the newly created ``.mop`` file.

    """

    mop_file = file_root + '.mop'
    mol = molecule.mol

    logger.info('Creating .mop file - {}.'.format(file_root))

    # Generate the mop file containing the MOPAC run info.
    with open(mop_file, 'w') as mop:
        # Line for the run info.
        mop.write(_mop_line(settings) + "\n")
        # Line with the name of the molecule.
        mop.write(file_root + "\n\n")

        # Write the structural info.
        for atom in mol.GetAtoms():
            atom_id = atom.GetIdx()
            symbol = atom.GetSymbol()
            x, y, z = mol.GetConformer().GetAtomPosition(atom_id)
            atom_info = f'{symbol}   {x}   +1  {y}   +1  {z}   +1 \n'
            mop.write(atom_info)

    return mop_file


def _extract_MOPAC_en(file_root):
    mopac_out = file_root + '.arc'

    with open(mopac_out) as outfile:
        target = "TOTAL ENERGY"
        energy_str = str([x for x in outfile.readlines() if target in x][0])
        energy_val = float(energy_str.split()[3])

    return energy_val


def _extract_MOPAC_dipole(file_root):
    mopac_out = file_root + '.arc'

    with open(mopac_out) as outfile:
        target = "DIPOLE"
        energy_str = str([x for x in outfile.readlines() if target in x][0])
        energy_val = float(energy_str.split()[2])

    return energy_val



from uuid import uuid4
import os
import subprocess as sp

from ...utilities import move_generated_macromodel_files
from .energy_calculators import EnergyCalculator, EnergyError


class MacroModelEnergy(EnergyCalculator):
    """
    Calculates the energy using MacroModel.

    Attributes
    ----------
    macromodel_path : :class:`str`
        The full path of the Schrodinger suite within the user's
        machine. For example, on a Linux machine this may be something
        like ``'/opt/schrodinger2017-2'``.

    output_dir : :class:`str`
        The name of the directory into which files generated during
        the optimization are written, if ``None`` then
        :func:`uuid.uuid4` is used.

    force_field : :class:`int`
        The number of the force field to be used.

    """

    def __init__(self,
                 macromodel_path,
                 output_dir=None,
                 force_field=16,
                 use_cache=False):
        """

        Parameters
        ----------
        macromodel_path : :class:`str`
            The full path of the Schrodinger suite within the user's
            machine. For example, on a Linux machine this may be
            something like ``'/opt/schrodinger2017-2'``.

        output_dir : :class:`str`, optional
            The name of the directory into which files generated during
            the optimization are written, if ``None`` then
            :func:`uuid.uuid4` is used.

        force_field : :class:`int`, optional
            The number of the force field to be used.

        use_cache : :class:`bool`, optional
            If ``True`` :meth:`energy` will not run twice on the same
            molecule and conformer, but will instead return the previously
            calculated value.

        """

        self.macromodel_path = macromodel_path
        self.output_dir = output_dir
        self.force_field = force_field
        super().__init__(use_cache=use_cache)

    def energy(self, mol, conformer=-1):
        """
        Calculates the energy of `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The :class:`.Molecule` whose energy is to be calculated.

        conformer : :class:`int`, optinal
            The conformer of `mol` to use.

        Returns
        -------
        :class:`float`
            The energy.

        Raises
        ------
        :class:`EnergyError`
            This exception is raised if no energy value is found in the
            MacroModel calculation's ``.log`` file. Likely due to a
            forcefield error.

        """

        # To prevent conflicts when running this function in parallel,
        # a temporary copy of the molecular structure file is made and
        # used for macromodel calculations.

        # Unique file name is generated by inserting a random int into
        # the file path.
        basename = str(uuid4().int)
        if self.output_dir is None:
            output_dir = basename

        tmp_file = f'{basename}.mol'
        mol.write(tmp_file, conformer)

        convrt_app = os.path.join(self.macromodel_path,
                                  'utilities',
                                  'structconvert')
        convrt_cmd = [convrt_app,
                      tmp_file,
                      f'{basename}.mae']
        sp.call(convrt_cmd, stdout=sp.PIPE, stderr=sp.PIPE)

        # Create an input file and run it.
        input_script = (
         "{0}.mae\n"
         "{0}-out.maegz\n"
         " MMOD       0      1      0      0     0.0000     "
         "0.0000     0.0000     0.0000\n"
         " FFLD{1:8}      1      0      0     1.0000     "
         "0.0000     0.0000     0.0000\n"
         " BGIN       0      0      0      0     0.0000     "
         "0.0000     0.0000     0.0000\n"
         " READ      -1      0      0      0     0.0000     "
         "0.0000     0.0000     0.0000\n"
         " ELST      -1      0      0      0     0.0000     "
         "0.0000     0.0000     0.0000\n"
         " WRIT       0      0      0      0     0.0000     "
         "0.0000     0.0000     0.0000\n"
         " END       0      0      0      0     0.0000     "
         "0.0000     0.0000     0.0000\n\n"
        ).format(basename, self.force_field)

        with open(f'{basename}.com', 'w') as f:
            f.write(input_script)

        cmd = [
            os.path.join(self.macromodel_path, 'bmin'),
            basename,
            "-WAIT",
            "-LOCAL"
        ]
        sp.call(cmd)

        # Check if the license was found. If not run the function
        # again.
        with open(f'{basename}.log', 'r') as f:
            log_content = f.read()

        if ('FATAL -96: Could not check out a license for mmlibs' in
           log_content):
            return self.energy(mol, conformer)

        # Read the .log file and return the energy.
        with open(f'{basename}.log', 'r') as f:
            for line in f:
                if "                   Total Energy =" in line:
                    eng = float(line.split()[-2].replace("=", ""))

        try:
            return eng
        except UnboundLocalError:
            raise EnergyError('MacroModel energy calculation failed.')

        finally:
            move_generated_macromodel_files(basename, output_dir)



from distutils.core import setup
import re
from os.path import join


def get_version():
    with open(join('stk', '__init__.py'), 'r') as f:
        content = f.read()
    p = re.compile(r'^__version__ = [\'"]([^\'\"]*)[\'"]', re.M)
    return p.search(content).group(1)


setup(name='stk',
      author='Lukas Turcani',
      author_email='lukasturcani93@gmail.com',
      url='https://www.github.com/lukasturcani/stk',
      version=get_version(),
      packages=[
          'stk',
          'stk.utilities',
          'stk.molecular',
          'stk.molecular.topologies',
          'stk.molecular.topologies.cage',
          'stk.calculators',
          'stk.calculators.electronic_property',
          'stk.calculators.energy',
          'stk.calculators.optimization'
      ],
      install_requires=[
          'networkx',
          'scipy',
          'matplotlib',
          'scikit-learn',
          'psutil',
          'pywindowx',
          'pandas',
          'seaborn'
       ],
      python_requires='>=3.6')



"""
Defines cage topologies of building blocks with 3 functional groups.

"""

import numpy as np
from scipy.spatial.distance import euclidean

from .base import NoLinkerCageTopology,  Vertex


class OnePlusOne(NoLinkerCageTopology):
    """
    A sandwich cage topology from tri-functionalised building blocks.

    """

    x = 1
    positions_A = [Vertex(x, 0., 0.),
                   Vertex(-x, 0., 0.)]
    a, b = positions_A
    connections = [(a, b)]

    a.edge_plane_normal = lambda scale: scale*np.array([1, 0, 0])
    b.edge_plane_normal = lambda scale: scale*np.array([-1, 0, 0])

    a.edge_centroid = lambda scale: scale*np.array([0, 0, 0])
    b.edge_centroid = lambda scale: scale*np.array([0, 0, 0])

    n_windows = 3
    n_window_types = 1

    def bonded_fgs(self, macro_mol):

        for position in self.positions_A:
            other_position = next(x for x in self.positions_A if
                                  x is not position)

            position.fg_position_pairs = [
                (fg, other_position) for fg in position.fgs
            ]

            for fg1, vertex in position.fg_position_pairs:
                # Get all the distances between the fg and the fgs
                # on the vertex. Store this information on the vertex.

                for fg2 in vertex.fgs:
                    c1 = macro_mol.atom_centroid(fg1.bonder_ids)
                    c2 = macro_mol.atom_centroid(fg2.bonder_ids)
                    distance = euclidean(c1, c2)
                    position.distances.append((distance, fg1, fg2))

        paired = set()
        for position in self.positions_A:
            for _, fg1, fg2 in sorted(position.distances):
                if fg1 in paired or fg2 in paired:
                    continue

                # Add the bond.
                yield fg1, fg2
                paired.add(fg1)
                paired.add(fg2)


class TwoPlusTwo(NoLinkerCageTopology):
    """
    Tetrahedral cage topology from tri-functionalised building blocks.

    """

    x = 1
    positions_A = [Vertex(x, 0, -x/np.sqrt(2)),
                   Vertex(-x, 0, -x/np.sqrt(2)),
                   Vertex(0, x, x/np.sqrt(2)),
                   Vertex(0, -x, x/np.sqrt(2))]

    a, b, c, d = positions_A

    for x in positions_A:
        old_normal = x.edge_plane_normal
        x.edge_plane_normal = lambda scale, a=old_normal: -1*a(scale)

    connections = [(a, b), (a, c), (a, d),
                   (b, c), (b, d),
                   (c, d)]

    n_windows = 4
    n_window_types = 1


class FourPlusFour(NoLinkerCageTopology):
    """
    A square cage topology from tri-functionalised building blocks.

    """

    x = 1
    positions_A = [Vertex(-x, x, -x),
                   Vertex(-x, -x, -x),
                   Vertex(x, x, -x),
                   Vertex(x, -x, -x),

                   Vertex(-x, x, x),
                   Vertex(-x, -x, x),
                   Vertex(x, x, x),
                   Vertex(x, -x, x)]

    a, b, c, d, e, f, g, h = positions_A

    connections = [(a, b), (a, c), (a, e), (b, d), (b, f), (c, g),
                   (c, d), (d, h), (e, g), (e, f), (f, h), (g, h)]

    n_windows = 6
    n_window_types = 1



"""
Defines cage topologies from 2 and 3 functionalized building blocks.

"""

import numpy as np

from .base import CageTopology, Vertex, Edge


class TwoPlusThree(CageTopology):
    """
    A cage topology from 2 and 3 functionalized building blocks.

    """
    positions_A = a, b = [Vertex(0, 0, 1), Vertex(0, 0, -1)]

    positions_B = alpha, beta, gamma = [Edge(a, b),
                                        Edge(a, b),
                                        Edge(a, b)]

    b.edge_plane_normal = (
        lambda scale, a=a: np.multiply(a.edge_plane_normal(scale), -1)
    )

    alpha.coord = np.array([-1, -0.5*np.sqrt(3), 0])
    beta.coord = np.array([1, -0.5*np.sqrt(3), 0])
    gamma.coord = np.array([0, 0.5*np.sqrt(3), 0])

    n_windows = 3
    n_window_types = 1


class FourPlusSix(CageTopology):
    """
    Defines the tetrahedral, 4+6, topology.

    This is a topology of cages where 4 building blocks are placed on
    vertices and 6 linkers are placed on the edges between them.

    """

    # Vertices of a tetrahdron so that origin is at the origin. Source:
    # http://tinyurl.com/lc262h8.
    positions_A = v0, v1, v2, v3 = [
                Vertex(0, 0, np.sqrt(6)/2),
                Vertex(-1, -np.sqrt(3)/3, -np.sqrt(6)/6),
                Vertex(1, -np.sqrt(3)/3, -np.sqrt(6)/6),
                Vertex(0, 2*np.sqrt(3)/3, -np.sqrt(6)/6)]

    positions_B = [Edge(v0, v1, (0, 1)), Edge(v0, v2, (0, 2)),
                   Edge(v0, v3, (0, 3)), Edge(v1, v2, (1, 2)),
                   Edge(v1, v3, (1, 3)), Edge(v2, v3, (2, 3))]

    n_windows = 4
    n_window_types = 1


class FourPlusSix2(CageTopology):
    """
    Defines the 4+6 topolgy which is not a tetrahedron.

    """

    positions_A = a, b, c, d = [Vertex(1, 0, 1),
                                Vertex(-1, 0, 1),
                                Vertex(1, 0, -1),
                                Vertex(-1, 0, -1)]

    positions_B = e1, e2, e3, e4, e5, e6 = [Edge(a, b, 'a', True),
                                            Edge(a, b, 'b', True),
                                            Edge(c, d, 'e', True),
                                            Edge(c, d, 'f', True),
                                            Edge(a, c, 'd'),
                                            Edge(b, d, 'c')]

    e1.coord = np.array([0, -1, 1])
    e2.coord = np.array([0, 1, 1])
    e3.coord = np.array([0, -1, -1])
    e4.coord = np.array([0, 1, -1])


class SixPlusNine(CageTopology):
    """
    A cage topology from 2 and 3 functionalized building blocks.

    """

    # source: http://eusebeia.dyndns.org/4d/prism3
    positions_A = [Vertex(-1, -1/np.sqrt(3), -1),
                   Vertex(-1, -1/np.sqrt(3), 1),
                   Vertex(1, -1/np.sqrt(3), -1),
                   Vertex(1, -1/np.sqrt(3), 1),
                   Vertex(0, 2/np.sqrt(3), -1),
                   Vertex(0, 2/np.sqrt(3), 1)]

    a, b, c, d, e, f = positions_A

    positions_B = [Edge(a, b),
                   Edge(a, c),
                   Edge(c, d),
                   Edge(b, d),
                   Edge(a, e),
                   Edge(c, e),
                   Edge(b, f),
                   Edge(d, f),
                   Edge(e, f)]

    n_windows = 5
    n_window_types = 1


class EightPlusTwelve(CageTopology):
    """
    A square topology from 2 and 3 functionalized building blocks.

    """

    positions_A = [Vertex(-1, 1, -1),
                   Vertex(-1, -1, -1),
                   Vertex(1, 1, -1),
                   Vertex(1, -1, -1),

                   Vertex(-1, 1, 1),
                   Vertex(-1, -1, 1),
                   Vertex(1, 1, 1),
                   Vertex(1, -1, 1)]

    a, b, c, d, e, f, g, h = positions_A

    positions_B = [Edge(a, c),
                   Edge(a, b),
                   Edge(b, d),
                   Edge(c, d),

                   Edge(e, g),
                   Edge(e, f),
                   Edge(f, h),
                   Edge(g, h),


                   Edge(a, e),
                   Edge(b, f),
                   Edge(c, g),
                   Edge(d, h)]

    n_windows = 6
    n_window_types = 1


class Dodecahedron(CageTopology):
    """
    A dodecahedron cage from 2 and 3 functionalized building blocks.

    """

    # Source: http://tinyurl.com/h2dl949
    phi = (1 + np.sqrt(5))/2
    x = 1.5
    positions_A = [Vertex(x*phi, 0.0, x/phi),
                   Vertex(x*-phi, 0.0, x/phi),
                   Vertex(x*-phi, 0.0, x/-phi),
                   Vertex(x*phi, 0.0, x/-phi),

                   Vertex(x/phi, x*phi, 0.0),
                   Vertex(x/phi, x*-phi, 0.0),
                   Vertex(x/-phi, x*-phi, 0.0),
                   Vertex(x/-phi, x*phi, 0.0),
                   Vertex(0.0, x/phi, x*phi),
                   Vertex(0.0, x/phi, x*-phi),
                   Vertex(0.0, x/-phi, x*-phi),
                   Vertex(0.0, x/-phi, x*phi),

                   Vertex(x, x, x),
                   Vertex(x, -x, x),
                   Vertex(-x, -x, x),
                   Vertex(-x, x, x),
                   Vertex(-x, x, -x),
                   Vertex(x, x, -x),
                   Vertex(x, -x, -x),
                   Vertex(-x, -x, -x)]

    A, B, C, D, E, F, G, *_ = positions_A
    H, I, J, K, L, M, N, O, P, Q, R, S, T = _

    positions_B = [
         Edge(A, N), Edge(A, M), Edge(A, D), Edge(B, O), Edge(B, P),
         Edge(B, C), Edge(C, T), Edge(C, Q), Edge(D, S), Edge(D, R),
         Edge(E, M), Edge(E, H), Edge(E, R), Edge(F, G), Edge(F, S),
         Edge(F, N), Edge(G, O), Edge(G, T), Edge(H, P), Edge(H, Q),
         Edge(I, L), Edge(I, M), Edge(I, P), Edge(J, K), Edge(J, R),
         Edge(J, Q), Edge(K, S), Edge(K, T), Edge(L, O), Edge(L, N)]

    n_windows = 12
    n_window_types = 1
